"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2018 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.field_mask_pb2
import google.protobuf.internal.containers
import google.protobuf.message
import buildgrid._protos.google.rpc.status_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Task(google.protobuf.message.Message):
    """DEPRECATED - use Lease.payload instead.
    A Task represents a unit of work. Its result and logs are defined as
    subresources.

    If all the `Any` fields are populated, this can be a very large message, and
    clients may not want the entire message returned on every call to every
    method. Such clients should request partial responses
    (https://cloud.google.com/apis/design/design_patterns#partial_response) and
    servers should implement partial responses in order to reduce unnecessry
    overhead.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class LogsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    LOGS_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The name of this task. Output only."""
    @property
    def description(self) -> google.protobuf.any_pb2.Any:
        """The actual task to perform. For example, this could be CommandTask to run a
        command line.
        """
    @property
    def logs(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """Handles to logs. The key is a human-readable name like `stdout`, and the
        handle is a resource name that can be passed to ByteStream or other
        accessors.

        An implementation may define some logs by default (like `stdout`), and may
        allow clients to add new logs via AddTaskLog.
        """
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        description: google.protobuf.any_pb2.Any | None = ...,
        logs: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["description", b"description"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["description", b"description", "logs", b"logs", "name", b"name"]) -> None: ...

global___Task = Task

@typing_extensions.final
class TaskResult(google.protobuf.message.Message):
    """DEPRECATED - use Lease.assignment_result instead.
    The result and metadata of the task.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    COMPLETE_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    OUTPUT_FIELD_NUMBER: builtins.int
    META_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The name of the task result; must be a name of a `Task` followed by
    `/result`.
    """
    complete: builtins.bool
    """The result may be updated several times; the client must only set
    `complete` to true to indicate that no further updates are allowed.
    If this is not true, the `status` field must not be examined since its zero
    value is equivalent to `OK`.

    Once a task is completed, it must not be updated with further results,
    though the implementation may choose to continue to receive logs.
    """
    @property
    def status(self) -> buildgrid._protos.google.rpc.status_pb2.Status:
        """The final status of the task itself. For example, if task.description
        included a timeout which was violated, status.code may be
        DEADLINE_EXCEEDED. This field can only be read if `complete` is true.
        """
    @property
    def output(self) -> google.protobuf.any_pb2.Any:
        """Any non-log output, such as output files and exit codes. See
        CommandResult as an example.
        """
    @property
    def meta(self) -> google.protobuf.any_pb2.Any:
        """Any information about how the command was executed, eg runtime. See
        CommandOverhead as an example.
        """
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        complete: builtins.bool = ...,
        status: buildgrid._protos.google.rpc.status_pb2.Status | None = ...,
        output: google.protobuf.any_pb2.Any | None = ...,
        meta: google.protobuf.any_pb2.Any | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["meta", b"meta", "output", b"output", "status", b"status"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["complete", b"complete", "meta", b"meta", "name", b"name", "output", b"output", "status", b"status"]) -> None: ...

global___TaskResult = TaskResult

@typing_extensions.final
class GetTaskRequest(google.protobuf.message.Message):
    """Request message for `GetTask`."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The task name."""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name"]) -> None: ...

global___GetTaskRequest = GetTaskRequest

@typing_extensions.final
class UpdateTaskResultRequest(google.protobuf.message.Message):
    """Request message for `UpdateTaskResult`."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    RESULT_FIELD_NUMBER: builtins.int
    UPDATE_MASK_FIELD_NUMBER: builtins.int
    SOURCE_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The task result name; must match `result.name`."""
    @property
    def result(self) -> global___TaskResult:
        """The result being updated."""
    @property
    def update_mask(self) -> google.protobuf.field_mask_pb2.FieldMask:
        """The fields within `result` that are specified."""
    source: builtins.str
    """If this is being updated by a bot from BotManager, the source should be
    bot.session_id. That way, if two bots accidentally get the same name, we'll
    know to reject updates from the older one.
    """
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        result: global___TaskResult | None = ...,
        update_mask: google.protobuf.field_mask_pb2.FieldMask | None = ...,
        source: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["result", b"result", "update_mask", b"update_mask"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name", "result", b"result", "source", b"source", "update_mask", b"update_mask"]) -> None: ...

global___UpdateTaskResultRequest = UpdateTaskResultRequest

@typing_extensions.final
class AddTaskLogRequest(google.protobuf.message.Message):
    """Request message for `AddTaskLog`."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    LOG_ID_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The name of the task that will own the new log."""
    log_id: builtins.str
    """The human-readable name of the log, like `stdout` or a relative file path."""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        log_id: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["log_id", b"log_id", "name", b"name"]) -> None: ...

global___AddTaskLogRequest = AddTaskLogRequest

@typing_extensions.final
class AddTaskLogResponse(google.protobuf.message.Message):
    """Response message for `AddTaskLog`."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    HANDLE_FIELD_NUMBER: builtins.int
    handle: builtins.str
    """The handle for the new log, as would be returned in Task.logs."""
    def __init__(
        self,
        *,
        handle: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["handle", b"handle"]) -> None: ...

global___AddTaskLogResponse = AddTaskLogResponse



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright (C) 2024 Bloomberg LP

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

 <http://www.apache.org/licenses/LICENSE-2.0>

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2
import builtins
import collections.abc
import buildgrid._protos.google.devtools.remoteworkers.v1test2.bots_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class ListWorkersRequest(google.protobuf.message.Message):
    """Request specifying which worker(s) to list"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INSTANCE_NAME_FIELD_NUMBER: builtins.int
    PAGE_SIZE_FIELD_NUMBER: builtins.int
    PAGE_FIELD_NUMBER: builtins.int
    WORKER_NAME_FIELD_NUMBER: builtins.int
    instance_name: builtins.str
    """The name of the BuildGrid instance to query workers in"""
    page_size: builtins.int
    """The maximum number of workers to return in a single response"""
    page: builtins.int
    """The page number to return when paginating"""
    worker_name: builtins.str
    """A string which is matched against worker names and IDs"""
    def __init__(
        self,
        *,
        instance_name: builtins.str = ...,
        page_size: builtins.int = ...,
        page: builtins.int = ...,
        worker_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["instance_name", b"instance_name", "page", b"page", "page_size", b"page_size", "worker_name", b"worker_name"]) -> None: ...

global___ListWorkersRequest = ListWorkersRequest

@typing_extensions.final
class WorkerOperation(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OPERATION_NAME_FIELD_NUMBER: builtins.int
    operation_name: builtins.str
    """The name of an operation being addressed by a worker"""
    def __init__(
        self,
        *,
        operation_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["operation_name", b"operation_name"]) -> None: ...

global___WorkerOperation = WorkerOperation

@typing_extensions.final
class RegisteredWorker(google.protobuf.message.Message):
    """A Worker message defines the internal tracked state for a bot connected
    via the Remote Workers API.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SESSION_NAME_FIELD_NUMBER: builtins.int
    WORKER_NAME_FIELD_NUMBER: builtins.int
    LAST_UPDATED_FIELD_NUMBER: builtins.int
    BOT_STATUS_FIELD_NUMBER: builtins.int
    EXPIRY_TIME_FIELD_NUMBER: builtins.int
    OPERATIONS_FIELD_NUMBER: builtins.int
    ACTION_DIGEST_FIELD_NUMBER: builtins.int
    session_name: builtins.str
    """The name of the worker's BotSession, generated by BuildGrid when the
    worker's session starts
    """
    worker_name: builtins.str
    """The ID of the worker's BotSession, specified by the worker client. This
    should be the is the same worker name that ends up in the
    ExecuteOperationMetadata
    """
    @property
    def last_updated(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """The time at which the worker last connected to BuildGrid"""
    bot_status: buildgrid._protos.google.devtools.remoteworkers.v1test2.bots_pb2.BotStatus.ValueType
    """The status of the worker when it last communicated"""
    @property
    def expiry_time(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """The time at which this worker will expire and any assigned work be retried"""
    @property
    def operations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___WorkerOperation]:
        """The list of operations assigned to this worker"""
    @property
    def action_digest(self) -> buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest:
        """The Action Digest for the assigned job, if any"""
    def __init__(
        self,
        *,
        session_name: builtins.str = ...,
        worker_name: builtins.str = ...,
        last_updated: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        bot_status: buildgrid._protos.google.devtools.remoteworkers.v1test2.bots_pb2.BotStatus.ValueType = ...,
        expiry_time: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        operations: collections.abc.Iterable[global___WorkerOperation] | None = ...,
        action_digest: buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["action_digest", b"action_digest", "expiry_time", b"expiry_time", "last_updated", b"last_updated"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["action_digest", b"action_digest", "bot_status", b"bot_status", "expiry_time", b"expiry_time", "last_updated", b"last_updated", "operations", b"operations", "session_name", b"session_name", "worker_name", b"worker_name"]) -> None: ...

global___RegisteredWorker = RegisteredWorker

@typing_extensions.final
class ListWorkersResponse(google.protobuf.message.Message):
    """Response to a ListWorkers request, containing the matching list of Workers"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PAGE_FIELD_NUMBER: builtins.int
    PAGE_SIZE_FIELD_NUMBER: builtins.int
    TOTAL_FIELD_NUMBER: builtins.int
    WORKERS_FIELD_NUMBER: builtins.int
    page: builtins.int
    """The page number this response contains when using the specified page_size"""
    page_size: builtins.int
    """The requested page size"""
    total: builtins.int
    """The total number of workers which matched the requested `worker_name`"""
    @property
    def workers(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___RegisteredWorker]:
        """A list of Workers which match the requested `worker_name`"""
    def __init__(
        self,
        *,
        page: builtins.int = ...,
        page_size: builtins.int = ...,
        total: builtins.int = ...,
        workers: collections.abc.Iterable[global___RegisteredWorker] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["page", b"page", "page_size", b"page_size", "total", b"total", "workers", b"workers"]) -> None: ...

global___ListWorkersResponse = ListWorkersResponse

@typing_extensions.final
class GetOperationFiltersRequest(google.protobuf.message.Message):
    """Request specifying an instance to get OperationFilters for"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INSTANCE_NAME_FIELD_NUMBER: builtins.int
    instance_name: builtins.str
    """The name of the BuildGrid instance to get OperationFilters for"""
    def __init__(
        self,
        *,
        instance_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["instance_name", b"instance_name"]) -> None: ...

global___GetOperationFiltersRequest = GetOperationFiltersRequest

@typing_extensions.final
class OperationFilter(google.protobuf.message.Message):
    """Specification for a key to filter Operations by in ListOperations requests

    The filters are in the form `KEY COMPARATOR VALUE`, where the value may or
    may not come from the `values` field of this message. The other fields are
    metadata to help describe the format of the filter.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    COMPARATORS_FIELD_NUMBER: builtins.int
    VALUES_FIELD_NUMBER: builtins.int
    key: builtins.str
    """The filter key, used in the ListOperations filter string"""
    name: builtins.str
    """The filter name, used just for readability"""
    type: builtins.str
    """The type of values this filter accepts; either `text` or `datetime`"""
    description: builtins.str
    """Description of this filter key"""
    @property
    def comparators(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """List of valid comparators/operators to use with this filter key"""
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Optional list of values that are valid for this filter. If empty, the
        filter is assumed to accept any values of the specified type.
        """
    def __init__(
        self,
        *,
        key: builtins.str = ...,
        name: builtins.str = ...,
        type: builtins.str = ...,
        description: builtins.str = ...,
        comparators: collections.abc.Iterable[builtins.str] | None = ...,
        values: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["comparators", b"comparators", "description", b"description", "key", b"key", "name", b"name", "type", b"type", "values", b"values"]) -> None: ...

global___OperationFilter = OperationFilter

@typing_extensions.final
class OperationFilters(google.protobuf.message.Message):
    """Response containing a list of OperationFilter specs"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILTERS_FIELD_NUMBER: builtins.int
    @property
    def filters(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___OperationFilter]:
        """A list of OperationFilter messages describing all available filter keys for
        ListOperations on the requested instance
        """
    def __init__(
        self,
        *,
        filters: collections.abc.Iterable[global___OperationFilter] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["filters", b"filters"]) -> None: ...

global___OperationFilters = OperationFilters



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""remove optionals in jobs and bots fields

Revision ID: fe3a61461feb
Revises: e83194af8292
Create Date: 2024-08-30 12:39:10.653746

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "fe3a61461feb"
down_revision = "e83194af8292"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("jobs") as batch_op:
        batch_op.alter_column("instance_name", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.alter_column("action", existing_type=postgresql.BYTEA(), nullable=False)
        batch_op.alter_column("queued_timestamp", existing_type=postgresql.TIMESTAMP(), nullable=False)
        batch_op.alter_column("assigned", existing_type=sa.BOOLEAN(), nullable=False)
        batch_op.alter_column("n_tries", existing_type=sa.INTEGER(), nullable=False)
        batch_op.alter_column("platform_requirements", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.alter_column("command", existing_type=sa.VARCHAR(), nullable=False)

    with op.batch_alter_table("bots") as batch_op:
        batch_op.alter_column("instance_name", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.alter_column("expiry_time", existing_type=postgresql.TIMESTAMP(), nullable=False)


def downgrade() -> None:
    with op.batch_alter_table("jobs") as batch_op:
        batch_op.alter_column("instance_name", existing_type=sa.VARCHAR(), nullable=True)
        batch_op.alter_column("action", existing_type=postgresql.BYTEA(), nullable=True)
        batch_op.alter_column("queued_timestamp", existing_type=postgresql.TIMESTAMP(), nullable=True)
        batch_op.alter_column("assigned", existing_type=sa.BOOLEAN(), nullable=True)
        batch_op.alter_column("n_tries", existing_type=sa.INTEGER(), nullable=True)
        batch_op.alter_column("platform_requirements", existing_type=sa.VARCHAR(), nullable=True)
        batch_op.alter_column("command", existing_type=sa.VARCHAR(), nullable=True)

    with op.batch_alter_table("bots") as batch_op:
        batch_op.alter_column("instance_name", existing_type=sa.VARCHAR(), nullable=True)
        batch_op.alter_column("expiry_time", existing_type=postgresql.TIMESTAMP(), nullable=True)



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
from typing import Any, Callable, Iterator, TypeVar, Union, cast

import grpc
from google.protobuf.message import Message

from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_counter_metric

from .time import _service_metadata

Func = TypeVar("Func", bound=Callable)  # type: ignore[type-arg]
_Message = Union[Iterator[Message], Message]


def network_io(f: Func) -> Func:
    @functools.wraps(f)
    def server_stream_wrapper(self: Any, message: _Message, context: grpc.ServicerContext) -> Iterator[Any]:
        input_bytes = 0
        if isinstance(message, Iterator):

            def stream(messages: Iterator[Message]) -> Iterator[Message]:
                nonlocal input_bytes
                for input_message in messages:
                    input_bytes += input_message.ByteSize()
                    yield input_message

            message = stream(message)
        else:
            input_bytes = message.ByteSize()

        output_bytes = 0
        try:
            for output_message in f(self, message, context):
                output_bytes += output_message.ByteSize()
                yield output_message
        finally:
            metadata = _service_metadata(self, message, context)
            publish_counter_metric(METRIC.RPC.INPUT_BYTES, input_bytes, **metadata)
            publish_counter_metric(METRIC.RPC.OUTPUT_BYTES, output_bytes, **metadata)

    @functools.wraps(f)
    def server_unary_wrapper(self: Any, message: _Message, context: grpc.ServicerContext) -> Any:
        input_bytes = 0
        if isinstance(message, Iterator):

            def stream(messages: Iterator[Message]) -> Iterator[Message]:
                nonlocal input_bytes
                for input_message in messages:
                    input_bytes += input_message.ByteSize()
                    yield input_message

            message = stream(message)
        else:
            input_bytes = message.ByteSize()

        output_bytes = 0
        try:
            output_message = f(self, message, context)
            output_bytes += output_message.ByteSize()
            return output_message
        finally:
            metadata = _service_metadata(self, message, context)
            publish_counter_metric(METRIC.RPC.INPUT_BYTES, input_bytes, **metadata)
            publish_counter_metric(METRIC.RPC.OUTPUT_BYTES, output_bytes, **metadata)

    if inspect.isgeneratorfunction(f):
        return cast(Func, server_stream_wrapper)
    return cast(Func, server_unary_wrapper)



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import uuid
from collections import namedtuple
from contextlib import contextmanager
from functools import partial
from io import SEEK_END, BytesIO
from operator import attrgetter
from typing import IO, BinaryIO, Dict, Generator, Iterator, List, Optional, Set, Tuple, TypeVar, Union, cast

import grpc
from google.protobuf.message import Message

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import (
    BatchReadBlobsRequest,
    BatchUpdateBlobsRequest,
    Digest,
    Directory,
    DirectoryNode,
    FileNode,
    GetTreeRequest,
    SymlinkNode,
    Tree,
)
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2_grpc import ContentAddressableStorageStub
from buildgrid._protos.google.bytestream.bytestream_pb2 import ReadRequest, WriteRequest
from buildgrid._protos.google.bytestream.bytestream_pb2_grpc import ByteStreamStub
from buildgrid._protos.google.rpc import code_pb2
from buildgrid.server.client.capabilities import CapabilitiesInterface
from buildgrid.server.client.retrier import GrpcRetrier
from buildgrid.server.exceptions import NotFoundError
from buildgrid.server.metadata import metadata_list
from buildgrid.server.settings import BATCH_REQUEST_SIZE_THRESHOLD, HASH, MAX_REQUEST_COUNT, MAX_REQUEST_SIZE
from buildgrid.server.types import MessageType
from buildgrid.server.utils.digests import create_digest

_FileRequest = namedtuple("_FileRequest", ["digest", "output_paths"])


def create_digest_from_file(file_obj: BinaryIO) -> Digest:
    """Computed the :obj:`Digest` of a file-like object.

    The :obj:`Digest` contains a hash of the file's contents and the size of
    those contents. This function only reads the content in chunks for hashing,
    so is safe to use on large files.

    Args:
        file_obj (BinaryIO): A file-like object of some kind.

    Returns:
        :obj:`Digest`: The :obj:`Digest` for the given file object.
    """
    digest = Digest()

    # Make sure we're hashing from the start of the file
    file_obj.seek(0)

    # Generate the file hash and keep track of the file size
    hasher = HASH()
    digest.size_bytes = 0
    for block in iter(partial(file_obj.read, 8192), b""):
        hasher.update(block)
        digest.size_bytes += len(block)
    digest.hash = hasher.hexdigest()

    # Return to the start of the file ready for future reads
    file_obj.seek(0)
    return digest


def merkle_tree_maker(directory_path: str) -> Iterator[Tuple[Union[FileNode, DirectoryNode], BinaryIO, str]]:
    """Walks a local folder tree, generating :obj:`FileNode` and
    :obj:`DirectoryNode`.

    Args:
        directory_path (str): absolute or relative path to a local directory.

    Yields:
        :obj:`Message`, bytes, str: a tutple of either a :obj:`FileNode` or
        :obj:`DirectoryNode` message, the corresponding blob and the
        corresponding node path.
    """
    directory_name = os.path.basename(directory_path)

    # Actual generator, yields recursively FileNodes and DirectoryNodes:
    def __merkle_tree_maker(directory_path: str, directory_name: str) -> Generator[
        Tuple[Union[FileNode, DirectoryNode], BinaryIO, str],
        None,
        Tuple[Union[FileNode, DirectoryNode], BinaryIO, str],
    ]:
        if not os.path.isabs(directory_path):
            directory_path = os.path.abspath(directory_path)

        directory = Directory()

        files, directories, symlinks = [], [], []
        for directory_entry in os.scandir(directory_path):
            node_name, node_path = directory_entry.name, directory_entry.path

            node: Union[FileNode, DirectoryNode]
            node_blob: BinaryIO
            if directory_entry.is_file(follow_symlinks=False):
                with open(directory_entry.path, "rb") as node_blob:
                    node_digest = create_digest_from_file(node_blob)

                    node = FileNode()
                    node.name = node_name
                    node.digest.CopyFrom(node_digest)
                    node.is_executable = os.access(node_path, os.X_OK)

                    files.append(node)

                    yield node, node_blob, node_path

            elif directory_entry.is_dir(follow_symlinks=False):
                node, node_blob, _ = yield from __merkle_tree_maker(node_path, node_name)

                directories.append(cast(DirectoryNode, node))

                yield node, node_blob, node_path

            # Create a SymlinkNode;
            elif os.path.islink(directory_entry.path):
                node_target = os.readlink(directory_entry.path)

                symlink_node = SymlinkNode()
                symlink_node.name = directory_entry.name
                symlink_node.target = node_target

                symlinks.append(symlink_node)

        files.sort(key=attrgetter("name"))
        directories.sort(key=attrgetter("name"))
        symlinks.sort(key=attrgetter("name"))

        directory.files.extend(files)
        directory.directories.extend(directories)
        directory.symlinks.extend(symlinks)

        node_data = directory.SerializeToString()
        node_digest = create_digest(node_data)

        dir_node = DirectoryNode()
        dir_node.name = directory_name
        dir_node.digest.CopyFrom(node_digest)

        return dir_node, BytesIO(node_data), directory_path

    node, node_blob, node_path = yield from __merkle_tree_maker(directory_path, directory_name)

    yield node, node_blob, node_path


class _CallCache:
    """Per remote grpc.StatusCode.UNIMPLEMENTED call cache."""

    __calls: Dict[grpc.Channel, Set[str]] = {}

    @classmethod
    def mark_unimplemented(cls, channel: grpc.Channel, name: str) -> None:
        if channel not in cls.__calls:
            cls.__calls[channel] = set()
        cls.__calls[channel].add(name)

    @classmethod
    def unimplemented(cls, channel: grpc.Channel, name: str) -> bool:
        if channel not in cls.__calls:
            return False
        return name in cls.__calls[channel]


class _CasBatchRequestSizesCache:
    """Cache that stores, for each remote, the limit of bytes that can
    be transferred using batches as well as a threshold that determines
    when a file can be fetched as part of a batch request.
    """

    __cas_max_batch_transfer_size: Dict[grpc.Channel, Dict[str, int]] = {}
    __cas_batch_request_size_threshold: Dict[grpc.Channel, Dict[str, int]] = {}

    @classmethod
    def max_effective_batch_size_bytes(cls, channel: grpc.Channel, instance_name: str) -> int:
        """Returns the maximum number of bytes that can be transferred
        using batch methods for the given remote.
        """
        if channel not in cls.__cas_max_batch_transfer_size:
            cls.__cas_max_batch_transfer_size[channel] = {}

        if instance_name not in cls.__cas_max_batch_transfer_size[channel]:
            max_batch_size = cls._get_server_max_batch_total_size_bytes(channel, instance_name)

            cls.__cas_max_batch_transfer_size[channel][instance_name] = max_batch_size

        return cls.__cas_max_batch_transfer_size[channel][instance_name]

    @classmethod
    def batch_request_size_threshold(cls, channel: grpc.Channel, instance_name: str) -> int:
        if channel not in cls.__cas_batch_request_size_threshold:
            cls.__cas_batch_request_size_threshold[channel] = {}

        if instance_name not in cls.__cas_batch_request_size_threshold[channel]:
            # Computing the threshold:
            max_batch_size = cls.max_effective_batch_size_bytes(channel, instance_name)
            threshold = int(BATCH_REQUEST_SIZE_THRESHOLD * max_batch_size)

            cls.__cas_batch_request_size_threshold[channel][instance_name] = threshold

        return cls.__cas_batch_request_size_threshold[channel][instance_name]

    @classmethod
    def _get_server_max_batch_total_size_bytes(cls, channel: grpc.Channel, instance_name: str) -> int:
        """Returns the maximum number of bytes that can be effectively
        transferred using batches, considering the limits imposed by
        the server's configuration and by gRPC.
        """
        try:
            capabilities_interface = CapabilitiesInterface(channel)
            server_capabilities = capabilities_interface.get_capabilities(instance_name)

            cache_capabilities = server_capabilities.cache_capabilities

            max_batch_total_size = cache_capabilities.max_batch_total_size_bytes
            # The server could set this value to 0 (no limit set).
            if max_batch_total_size:
                return min(max_batch_total_size, MAX_REQUEST_SIZE)
        except ConnectionError:
            pass

        return MAX_REQUEST_SIZE


T = TypeVar("T", bound=MessageType)


class Downloader:
    """Remote CAS files, directories and messages download helper.

    The :class:`Downloader` class comes with a generator factory function that
    can be used together with the `with` statement for context management::

        from buildgrid.server.client.cas import download

        with download(channel, instance='build') as downloader:
            downloader.get_message(message_digest)
    """

    def __init__(
        self,
        channel: grpc.Channel,
        instance: Optional[str] = None,
        retries: int = 0,
        max_backoff: int = 64,
        should_backoff: bool = True,
    ):
        """Initializes a new :class:`Downloader` instance.

        Args:
            channel (grpc.Channel): A gRPC channel to the CAS endpoint.
            instance (str, optional): the targeted instance's name.
        """
        self.channel = channel

        self.instance_name = instance or ""

        self._grpc_retrier = GrpcRetrier(retries=retries, max_backoff=max_backoff, should_backoff=should_backoff)

        self.__bytestream_stub: Optional[ByteStreamStub] = ByteStreamStub(self.channel)
        self.__cas_stub: Optional[ContentAddressableStorageStub] = ContentAddressableStorageStub(self.channel)

        self.__file_requests: Dict[str, _FileRequest] = {}
        self.__file_request_count = 0
        self.__file_request_size = 0
        self.__file_response_size = 0

    # --- Public API ---

    def get_blob(self, digest: Digest) -> Optional[bytearray]:
        """Retrieves a blob from the remote CAS server.

        Args:
            digest (:obj:`Digest`): the blob's digest to fetch.

        Returns:
            bytearray: the fetched blob data or None if not found.
        """
        try:
            blob = self._grpc_retrier.retry(self._fetch_blob, digest)
        except NotFoundError:
            return None

        return blob

    def get_blobs(self, digests: List[Digest]) -> List[bytes]:
        """Retrieves a list of blobs from the remote CAS server.

        Args:
            digests (list): list of :obj:`Digest`s for the blobs to fetch.

        Returns:
            list: the fetched blob data list.

        Raises:
            NotFoundError: if a blob is not present in the remote CAS server.
        """
        # _fetch_blob_batch returns (data, digest) pairs.
        # We only want the data.
        return [result[0] for result in self._grpc_retrier.retry(self._fetch_blob_batch, digests)]

    def get_available_blobs(self, digests: List[Digest]) -> List[Tuple[bytes, Digest]]:
        """Retrieves a list of blobs from the remote CAS server.

        Skips blobs not found on the server without raising an error.

        Args:
            digests (list): list of :obj:`Digest`s for the blobs to fetch.

        Returns:
            list: the fetched blob data list as (data, digest) pairs
        """
        return self._grpc_retrier.retry(self._fetch_blob_batch, digests, skip_unavailable=True)

    def get_message(self, digest: Digest, message: T) -> T:
        """Retrieves a :obj:`Message` from the remote CAS server.

        Args:
            digest (:obj:`Digest`): the message's digest to fetch.
            message (:obj:`Message`): an empty message to fill.

        Returns:
            :obj:`Message`: `message` filled or emptied if not found.
        """
        try:
            message_blob = self._grpc_retrier.retry(self._fetch_blob, digest)
        except NotFoundError:
            message_blob = None

        if message_blob is not None:
            message.ParseFromString(bytes(message_blob))
        else:
            message.Clear()

        return message

    def get_messages(self, digests: List[Digest], messages: List[Message]) -> List[Message]:
        """Retrieves a list of :obj:`Message`s from the remote CAS server.

        Note:
            The `digests` and `messages` list **must** contain the same number
            of elements.

        Args:
            digests (list):  list of :obj:`Digest`s for the messages to fetch.
            messages (list): list of empty :obj:`Message`s to fill.

        Returns:
            list: the fetched and filled message list.
        """
        assert len(digests) == len(messages)

        # The individual empty messages might be of differing types, so we need
        # to set up a mapping
        digest_message_map = {digest.hash: message for (digest, message) in zip(digests, messages)}

        batch_response = self._grpc_retrier.retry(self._fetch_blob_batch, digests)

        messages = []
        for message_blob, message_digest in batch_response:
            message = digest_message_map[message_digest.hash]
            message.ParseFromString(message_blob)
            messages.append(message)

        return messages

    def download_file(self, digest: Digest, file_path: str, is_executable: bool = False, queue: bool = True) -> None:
        """Retrieves a file from the remote CAS server.

        If queuing is allowed (`queue=True`), the download request **may** be
        defer. An explicit call to :func:`~flush` can force the request to be
        send immediately (along with the rest of the queued batch).

        Args:
            digest (:obj:`Digest`): the file's digest to fetch.
            file_path (str): absolute or relative path to the local file to write.
            is_executable (bool): whether the file is executable or not.
            queue (bool, optional): whether or not the download request may be
                queued and submitted as part of a batch upload request. Defaults
                to True.

        Raises:
            NotFoundError: if `digest` is not present in the remote CAS server.
            OSError: if `file_path` does not exist or is not readable.
        """
        if not os.path.isabs(file_path):
            file_path = os.path.abspath(file_path)

        if not queue or digest.size_bytes > self._queueable_file_size_threshold():
            self._grpc_retrier.retry(self._fetch_file, digest, file_path, is_executable)
        else:
            self._queue_file(digest, file_path, is_executable=is_executable)

    def download_directory(self, digest: Digest, directory_path: str) -> None:
        """Retrieves a :obj:`Directory` from the remote CAS server.

        Args:
            digest (:obj:`Digest`): the directory's digest to fetch.
            directory_path (str): the path to download to

        Raises:
            NotFoundError: if `digest` is not present in the remote CAS server.
            FileExistsError: if `directory_path` already contains parts of their
                fetched directory's content.
        """
        if not os.path.isabs(directory_path):
            directory_path = os.path.abspath(directory_path)

        self._grpc_retrier.retry(self._fetch_directory, digest, directory_path)

    def flush(self) -> None:
        """Ensures any queued request gets sent."""
        if self.__file_requests:
            self._grpc_retrier.retry(self._fetch_file_batch, self.__file_requests)

            self.__file_requests.clear()
            self.__file_request_count = 0
            self.__file_request_size = 0
            self.__file_response_size = 0

    def close(self) -> None:
        """Closes the underlying connection stubs.

        Note:
            This will always send pending requests before closing connections,
            if any.
        """
        self.flush()

        self.__bytestream_stub = None
        self.__cas_stub = None

    # --- Private API ---

    def _fetch_blob(self, digest: Digest) -> bytearray:
        """Fetches a blob using ByteStream.Read()"""

        assert self.__bytestream_stub, "Downloader used after closing"

        if self.instance_name:
            resource_name = "/".join([self.instance_name, "blobs", digest.hash, str(digest.size_bytes)])
        else:
            resource_name = "/".join(["blobs", digest.hash, str(digest.size_bytes)])

        read_blob = bytearray()
        read_request = ReadRequest()
        read_request.resource_name = resource_name
        read_request.read_offset = 0

        for read_response in self.__bytestream_stub.Read(read_request, metadata=metadata_list()):
            read_blob += read_response.data

        assert len(read_blob) == digest.size_bytes
        return read_blob

    def _fetch_blob_batch(
        self, digests: List[Digest], *, skip_unavailable: bool = False
    ) -> List[Tuple[bytes, Digest]]:
        """Fetches blobs using ContentAddressableStorage.BatchReadBlobs()
        Returns (data, digest) pairs"""

        assert self.__cas_stub, "Downloader used after closing"

        batch_fetched = False
        read_blobs = []

        # First, try BatchReadBlobs(), if not already known not being implemented:
        if not _CallCache.unimplemented(self.channel, "BatchReadBlobs"):
            batch_request = BatchReadBlobsRequest()
            batch_request.digests.extend(digests)
            if self.instance_name is not None:
                batch_request.instance_name = self.instance_name

            try:
                batch_response = self.__cas_stub.BatchReadBlobs(batch_request, metadata=metadata_list())

                for response in batch_response.responses:
                    assert response.digest in digests

                    if response.status.code == code_pb2.OK:
                        read_blobs.append((response.data, response.digest))
                    elif response.status.code == code_pb2.NOT_FOUND:
                        if not skip_unavailable:
                            raise NotFoundError("Requested blob does not exist on the remote.")
                    else:
                        raise ConnectionError("Error in CAS reply while fetching blob.")

                batch_fetched = True

            except grpc.RpcError as e:
                status_code = e.code()
                if status_code == grpc.StatusCode.UNIMPLEMENTED:
                    _CallCache.mark_unimplemented(self.channel, "BatchReadBlobs")
                elif status_code == grpc.StatusCode.INVALID_ARGUMENT:
                    read_blobs.clear()
                else:
                    raise

        # Fallback to Read() if no BatchReadBlobs():
        if not batch_fetched:
            for digest in digests:
                blob = self._grpc_retrier.retry(self._fetch_blob, digest)
                read_blobs.append((blob, digest))

        return read_blobs

    def _fetch_file(self, digest: Digest, file_path: str, is_executable: bool = False) -> None:
        """Fetches a file using ByteStream.Read()"""

        assert self.__bytestream_stub, "Downloader used after closing"

        if self.instance_name:
            resource_name = "/".join([self.instance_name, "blobs", digest.hash, str(digest.size_bytes)])
        else:
            resource_name = "/".join(["blobs", digest.hash, str(digest.size_bytes)])

        os.makedirs(os.path.dirname(file_path), exist_ok=True)

        read_request = ReadRequest()
        read_request.resource_name = resource_name
        read_request.read_offset = 0

        with open(file_path, "wb") as byte_file:
            for read_response in self.__bytestream_stub.Read(read_request, metadata=metadata_list()):
                byte_file.write(read_response.data)

            assert byte_file.tell() == digest.size_bytes

        if is_executable:
            os.chmod(file_path, 0o755)  # rwxr-xr-x / 755

    def _queue_file(self, digest: Digest, file_path: str, is_executable: bool = False) -> None:
        """Queues a file for later batch download"""
        batch_size_limit = self._max_effective_batch_size_bytes()

        if self.__file_request_size + digest.ByteSize() > batch_size_limit:
            self.flush()
        elif self.__file_response_size + digest.size_bytes > batch_size_limit:
            self.flush()
        elif self.__file_request_count >= MAX_REQUEST_COUNT:
            self.flush()

        output_path = (file_path, is_executable)

        # When queueing a file we take into account the cases where
        # we might want to download the same digest to different paths.
        if digest.hash not in self.__file_requests:
            request = _FileRequest(digest=digest, output_paths=[output_path])
            self.__file_requests[digest.hash] = request

            self.__file_request_count += 1
            self.__file_request_size += digest.ByteSize()
            self.__file_response_size += digest.size_bytes
        else:
            # We already have that hash queued; we'll fetch the blob
            # once and write copies of it:
            self.__file_requests[digest.hash].output_paths.append(output_path)

    def _fetch_file_batch(self, requests: Dict[str, _FileRequest]) -> None:
        """Sends queued data using ContentAddressableStorage.BatchReadBlobs().

        Takes a dictionary (digest.hash, _FileRequest) as input.
        """
        batch_digests = [request.digest for request in requests.values()]
        batch_responses = self._fetch_blob_batch(batch_digests)

        for file_blob, file_digest in batch_responses:
            output_paths = requests[file_digest.hash].output_paths

            for file_path, is_executable in output_paths:
                os.makedirs(os.path.dirname(file_path), exist_ok=True)

                with open(file_path, "wb") as byte_file:
                    byte_file.write(file_blob)

                if is_executable:
                    os.chmod(file_path, 0o755)  # rwxr-xr-x / 755

    def _fetch_directory(self, digest: Digest, directory_path: str) -> None:
        """Fetches a file using ByteStream.GetTree()"""
        # Better fail early if the local root path cannot be created:

        assert self.__cas_stub, "Downloader used after closing"

        os.makedirs(directory_path, exist_ok=True)

        directories = {}
        directory_fetched = False
        # First, try GetTree() if not known to be unimplemented yet:
        if not _CallCache.unimplemented(self.channel, "GetTree"):
            tree_request = GetTreeRequest()
            tree_request.root_digest.CopyFrom(digest)
            tree_request.page_size = MAX_REQUEST_COUNT
            if self.instance_name is not None:
                tree_request.instance_name = self.instance_name

            try:
                for tree_response in self.__cas_stub.GetTree(tree_request):
                    for directory in tree_response.directories:
                        directory_blob = directory.SerializeToString()
                        directory_hash = HASH(directory_blob).hexdigest()

                        directories[directory_hash] = directory

                assert digest.hash in directories

                directory = directories[digest.hash]
                self._write_directory(directory, directory_path, directories=directories)

                directory_fetched = True
            except grpc.RpcError as e:
                status_code = e.code()
                if status_code == grpc.StatusCode.UNIMPLEMENTED:
                    _CallCache.mark_unimplemented(self.channel, "GetTree")

                else:
                    raise

        # If no GetTree(), _write_directory() will use BatchReadBlobs()
        # if available or Read() if not.
        if not directory_fetched:
            directory = Directory()
            directory.ParseFromString(self._grpc_retrier.retry(self._fetch_blob, digest))

            self._write_directory(directory, directory_path)

    def _write_directory(
        self, root_directory: Directory, root_path: str, directories: Optional[Dict[str, Directory]] = None
    ) -> None:
        """Generates a local directory structure"""
        os.makedirs(root_path, exist_ok=True)
        self._write_directory_recursively(root_directory, root_path, directories=None)

    def _write_directory_recursively(
        self, root_directory: Directory, root_path: str, directories: Optional[Dict[str, Directory]] = None
    ) -> None:
        """Generate local directory recursively"""
        # i) Files:
        for file_node in root_directory.files:
            file_path = os.path.join(root_path, file_node.name)

            if os.path.lexists(file_path):
                raise FileExistsError(f"'{file_path}' already exists")

            self.download_file(file_node.digest, file_path, is_executable=file_node.is_executable)
        self.flush()

        # ii) Directories:
        pending_directory_digests = []
        pending_directory_paths = {}
        for directory_node in root_directory.directories:
            directory_hash = directory_node.digest.hash

            # FIXME: Guard against ../
            directory_path = os.path.join(root_path, directory_node.name)
            os.mkdir(directory_path)

            if directories and directory_node.digest.hash in directories:
                # We already have the directory; just write it:
                directory = directories[directory_hash]

                self._write_directory_recursively(directory, directory_path, directories=directories)
            else:
                # Gather all the directories that we need to get to
                # try fetching them in a single batch request:
                if directory_hash not in pending_directory_paths:
                    pending_directory_digests.append(directory_node.digest)
                    pending_directory_paths[directory_hash] = [directory_path]
                else:
                    pending_directory_paths[directory_hash].append(directory_path)

        if pending_directory_paths:
            fetched_blobs = self._grpc_retrier.retry(self._fetch_blob_batch, pending_directory_digests)

            for directory_blob, directory_digest in fetched_blobs:
                directory = Directory()
                directory.ParseFromString(directory_blob)

                # Assuming that the server might not return the blobs in
                # the same order than they were asked for, we read
                # the hashes of the returned blobs:
                # Guarantees for the reply orderings might change in
                # the specification at some point.
                # See: github.com/bazelbuild/remote-apis/issues/52

                for directory_path in pending_directory_paths[directory_digest.hash]:
                    self._write_directory(directory, directory_path, directories=directories)

        # iii) Symlinks:
        for symlink_node in root_directory.symlinks:
            symlink_path = os.path.join(root_path, symlink_node.name)
            os.symlink(symlink_node.target, symlink_path)

    def _max_effective_batch_size_bytes(self) -> int:
        """Returns the effective maximum number of bytes that can be
        transferred using batches, considering gRPC maximum message size.
        """
        return _CasBatchRequestSizesCache.max_effective_batch_size_bytes(self.channel, self.instance_name)

    def _queueable_file_size_threshold(self) -> int:
        """Returns the size limit up until which files can be queued to
        be requested in a batch.
        """
        return _CasBatchRequestSizesCache.batch_request_size_threshold(self.channel, self.instance_name)


@contextmanager
def download(
    channel: grpc.Channel,
    instance: Optional[str] = None,
    u_uid: Optional[str] = None,
    retries: int = 0,
    max_backoff: int = 64,
    should_backoff: bool = True,
) -> Iterator[Downloader]:
    """Context manager generator for the :class:`Downloader` class."""
    downloader = Downloader(
        channel, instance=instance, retries=retries, max_backoff=max_backoff, should_backoff=should_backoff
    )
    try:
        yield downloader
    finally:
        downloader.close()


class Uploader:
    """Remote CAS files, directories and messages upload helper.

    The :class:`Uploader` class comes with a generator factory function that can
    be used together with the `with` statement for context management::

        from buildgrid.server.client.cas import upload

        with upload(channel, instance='build') as uploader:
            uploader.upload_file('/path/to/local/file')
    """

    def __init__(
        self,
        channel: grpc.Channel,
        instance: Optional[str] = None,
        u_uid: Optional[str] = None,
        retries: int = 0,
        max_backoff: int = 64,
        should_backoff: bool = True,
    ):
        """Initializes a new :class:`Uploader` instance.

        Args:
            channel (grpc.Channel): A gRPC channel to the CAS endpoint.
            instance (str, optional): the targeted instance's name.
            u_uid (str, optional): a UUID for CAS transactions.
        """
        self.channel = channel

        self.instance_name = instance or ""
        if u_uid is not None:
            self.u_uid = u_uid
        else:
            self.u_uid = str(uuid.uuid4())

        self._grpc_retrier = GrpcRetrier(retries=retries, max_backoff=max_backoff, should_backoff=should_backoff)

        self.__bytestream_stub: Optional[ByteStreamStub] = ByteStreamStub(self.channel)
        self.__cas_stub: Optional[ContentAddressableStorageStub] = ContentAddressableStorageStub(self.channel)

        self.__requests: Dict[str, Tuple[bytes, Digest]] = {}
        self.__request_count = 0
        self.__request_size = 0

    # --- Public API ---

    def put_blob(
        self, blob: IO[bytes], digest: Optional[Digest] = None, queue: bool = False, length: Optional[int] = None
    ) -> Digest:
        """Stores a blob into the remote CAS server.

        If queuing is allowed (`queue=True`), the upload request **may** be
        defer. An explicit call to :func:`~flush` can force the request to be
        send immediately (along with the rest of the queued batch).

        The caller should set at least one of ``digest`` or ``length`` to
        allow the uploader to skip determining the size of the blob using
        multiple seeks.

        Args:
            blob (IO[bytes]): a file-like object containing the blob.
            digest (:obj:`Digest`, optional): the blob's digest.
            queue (bool, optional): whether or not the upload request may be
                queued and submitted as part of a batch upload request. Defaults
                to False.
            length (int, optional): The size of the blob, in bytes. If ``digest``
                is also set, this is ignored in favour of ``digest.size_bytes``.

        Returns:
            :obj:`Digest`: the sent blob's digest.
        """
        if digest is not None:
            length = digest.size_bytes
        elif length is None:
            # If neither the digest or the length were set, fall back to
            # seeking to the end of the object to find the length
            blob.seek(0, SEEK_END)
            length = blob.tell()
            blob.seek(0)

        if not queue or length > self._queueable_file_size_threshold():
            blob_digest = self._grpc_retrier.retry(self._send_blob, blob, digest)
        else:
            blob_digest = self._queue_blob(blob.read(), digest=digest)

        return blob_digest

    def put_message(self, message: Message, digest: Optional[Digest] = None, queue: bool = False) -> Digest:
        """Stores a message into the remote CAS server.

        If queuing is allowed (`queue=True`), the upload request **may** be
        defer. An explicit call to :func:`~flush` can force the request to be
        send immediately (along with the rest of the queued batch).

        Args:
            message (:obj:`Message`): the message object.
            digest (:obj:`Digest`, optional): the message's digest.
            queue (bool, optional): whether or not the upload request may be
                queued and submitted as part of a batch upload request. Defaults
                to False.

        Returns:
            :obj:`Digest`: the sent message's digest.
        """
        message_blob = message.SerializeToString()

        if not queue or len(message_blob) > self._queueable_file_size_threshold():
            message_digest = self._grpc_retrier.retry(self._send_blob, BytesIO(message_blob), digest)
        else:
            message_digest = self._queue_blob(message_blob, digest=digest)

        return message_digest

    def upload_file(self, file_path: str, queue: bool = True) -> Digest:
        """Stores a local file into the remote CAS storage.

        If queuing is allowed (`queue=True`), the upload request **may** be
        defer. An explicit call to :func:`~flush` can force the request to be
        send immediately (allong with the rest of the queued batch).

        Args:
            file_path (str): absolute or relative path to a local file.
            queue (bool, optional): whether or not the upload request may be
                queued and submitted as part of a batch upload request. Defaults
                to True.

        Returns:
            :obj:`Digest`: The digest of the file's content.

        Raises:
            FileNotFoundError: If `file_path` does not exist.
            PermissionError: If `file_path` is not readable.
        """
        if not os.path.isabs(file_path):
            file_path = os.path.abspath(file_path)

        with open(file_path, "rb") as file_object:
            if not queue or os.path.getsize(file_path) > self._queueable_file_size_threshold():
                file_digest = self._grpc_retrier.retry(self._send_blob, file_object)
            else:
                file_digest = self._queue_blob(file_object.read())

        return file_digest

    def upload_directory(self, directory_path: str, queue: bool = True) -> Digest:
        """Stores a local folder into the remote CAS storage.

        If queuing is allowed (`queue=True`), the upload request **may** be
        defer. An explicit call to :func:`~flush` can force the request to be
        send immediately (allong with the rest of the queued batch).

        Args:
            directory_path (str): absolute or relative path to a local folder.
            queue (bool, optional): wheter or not the upload requests may be
                queued and submitted as part of a batch upload request. Defaults
                to True.

        Returns:
            :obj:`Digest`: The digest of the top :obj:`Directory`.

        Raises:
            FileNotFoundError: If `directory_path` does not exist.
            PermissionError: If `directory_path` is not readable.
        """
        if not os.path.isabs(directory_path):
            directory_path = os.path.abspath(directory_path)

        if not queue:
            for node, blob, _ in merkle_tree_maker(directory_path):
                if node.DESCRIPTOR is DirectoryNode.DESCRIPTOR:
                    last_directory_node = node

                self._grpc_retrier.retry(self._send_blob, blob, node.digest)

        else:
            for node, blob, _ in merkle_tree_maker(directory_path):
                if node.DESCRIPTOR is DirectoryNode.DESCRIPTOR:
                    last_directory_node = node

                if node.digest.size_bytes > self._queueable_file_size_threshold():
                    self._grpc_retrier.retry(self._send_blob, blob, node.digest)
                else:
                    self._queue_blob(blob.read(), digest=node.digest)

        return last_directory_node.digest

    def upload_tree(self, directory_path: str, queue: bool = True) -> Digest:
        """Stores a local folder into the remote CAS storage as a :obj:`Tree`.

        If queuing is allowed (`queue=True`), the upload request **may** be
        defer. An explicit call to :func:`~flush` can force the request to be
        send immediately (allong with the rest of the queued batch).

        Args:
            directory_path (str): absolute or relative path to a local folder.
            queue (bool, optional): wheter or not the upload requests may be
                queued and submitted as part of a batch upload request. Defaults
                to True.

        Returns:
            :obj:`Digest`: The digest of the :obj:`Tree`.

        Raises:
            FileNotFoundError: If `directory_path` does not exist.
            PermissionError: If `directory_path` is not readable.
        """
        if not os.path.isabs(directory_path):
            directory_path = os.path.abspath(directory_path)

        directories = []

        if not queue:
            for node, blob, _ in merkle_tree_maker(directory_path):
                if node.DESCRIPTOR is DirectoryNode.DESCRIPTOR:
                    # TODO: Get the Directory object from merkle_tree_maker():
                    directory = Directory()
                    directory.ParseFromString(blob.read())
                    directories.append(directory)

                self._grpc_retrier.retry(self._send_blob, blob, node.digest)

        else:
            for node, blob, _ in merkle_tree_maker(directory_path):
                if node.DESCRIPTOR is DirectoryNode.DESCRIPTOR:
                    # TODO: Get the Directory object from merkle_tree_maker():
                    directory = Directory()
                    directory.ParseFromString(blob.read())
                    directories.append(directory)

                if node.digest.size_bytes > self._queueable_file_size_threshold():
                    self._grpc_retrier.retry(self._send_blob, blob, node.digest)
                else:
                    self._queue_blob(blob.read(), digest=node.digest)

        tree = Tree()
        tree.root.CopyFrom(directories[-1])
        tree.children.extend(directories[:-1])

        return self.put_message(tree, queue=queue)

    def flush(self) -> None:
        """Ensures any queued request gets sent."""
        if self.__requests:
            self._grpc_retrier.retry(self._send_blob_batch, self.__requests)

            self.__requests.clear()
            self.__request_count = 0
            self.__request_size = 0

    def close(self) -> None:
        """Closes the underlying connection stubs.

        Note:
            This will always send pending requests before closing connections,
            if any.
        """
        self.flush()

        self.__bytestream_stub = None
        self.__cas_stub = None

    # --- Private API ---

    def _send_blob(self, blob: BinaryIO, digest: Optional[Digest] = None) -> Digest:
        """Sends a memory block using ByteStream.Write()"""

        assert self.__bytestream_stub, "Uploader used after closing"

        blob.seek(0)
        blob_digest = Digest()
        if digest is not None:
            blob_digest.CopyFrom(digest)
        else:
            blob_digest = create_digest_from_file(blob)

        if self.instance_name:
            resource_name = "/".join(
                [self.instance_name, "uploads", self.u_uid, "blobs", blob_digest.hash, str(blob_digest.size_bytes)]
            )
        else:
            resource_name = "/".join(["uploads", self.u_uid, "blobs", blob_digest.hash, str(blob_digest.size_bytes)])

        def __write_request_stream(resource: str, content: BinaryIO) -> Iterator[WriteRequest]:
            offset = 0
            finished = False
            remaining = blob_digest.size_bytes - offset
            while not finished:
                chunk_size = min(remaining, MAX_REQUEST_SIZE)
                remaining -= chunk_size

                request = WriteRequest()
                request.resource_name = resource
                request.data = content.read(chunk_size)
                request.write_offset = offset
                request.finish_write = remaining <= 0

                yield request

                offset += chunk_size
                finished = request.finish_write

        write_requests = __write_request_stream(resource_name, blob)

        write_response = self.__bytestream_stub.Write(write_requests, metadata=metadata_list())

        assert write_response.committed_size == blob_digest.size_bytes

        return blob_digest

    def _queue_blob(self, blob: bytes, digest: Optional[Digest] = None) -> Digest:
        """Queues a memory block for later batch upload"""
        blob_digest = Digest()
        if digest is not None:
            blob_digest.CopyFrom(digest)
        else:
            blob_digest.hash = HASH(blob).hexdigest()
            blob_digest.size_bytes = len(blob)

        # If we are here queueing a file we know that its size is
        # smaller than gRPC's message size limit.
        # We'll make a single batch request as big as the server allows.
        batch_size_limit = self._max_effective_batch_size_bytes()

        if self.__request_size + blob_digest.size_bytes > batch_size_limit:
            self.flush()
        elif self.__request_count >= MAX_REQUEST_COUNT:
            self.flush()

        self.__requests[blob_digest.hash] = (blob, blob_digest)
        self.__request_count += 1
        self.__request_size += blob_digest.size_bytes

        return blob_digest

    def _send_blob_batch(self, batch: Dict[str, Tuple[bytes, Digest]]) -> List[Digest]:
        """Sends queued data using ContentAddressableStorage.BatchUpdateBlobs()"""

        assert self.__cas_stub, "Uploader used after closing"

        batch_fetched = False
        written_digests = []

        # First, try BatchUpdateBlobs(), if not already known not being implemented:
        if not _CallCache.unimplemented(self.channel, "BatchUpdateBlobs"):
            batch_request = BatchUpdateBlobsRequest()
            if self.instance_name is not None:
                batch_request.instance_name = self.instance_name

            for blob, digest in batch.values():
                request = batch_request.requests.add()
                request.digest.CopyFrom(digest)
                request.data = blob

            try:
                batch_response = self.__cas_stub.BatchUpdateBlobs(batch_request, metadata=metadata_list())

                for response in batch_response.responses:
                    assert response.digest.hash in batch

                    written_digests.append(response.digest)
                    if response.status.code != code_pb2.OK:
                        response.digest.Clear()

                batch_fetched = True

            except grpc.RpcError as e:
                status_code = e.code()
                if status_code == grpc.StatusCode.UNIMPLEMENTED:
                    _CallCache.mark_unimplemented(self.channel, "BatchUpdateBlobs")

                elif status_code == grpc.StatusCode.INVALID_ARGUMENT:
                    written_digests.clear()
                    batch_fetched = False

                else:
                    raise

        # Fallback to Write() if no BatchUpdateBlobs():
        if not batch_fetched:
            for blob, digest in batch.values():
                written_digests.append(self._send_blob(BytesIO(blob)))

        return written_digests

    def _max_effective_batch_size_bytes(self) -> int:
        """Returns the effective maximum number of bytes that can be
        transferred using batches, considering gRPC maximum message size.
        """
        return _CasBatchRequestSizesCache.max_effective_batch_size_bytes(self.channel, self.instance_name)

    def _queueable_file_size_threshold(self) -> int:
        """Returns the size limit up until which files can be queued to
        be requested in a batch.
        """
        return _CasBatchRequestSizesCache.batch_request_size_threshold(self.channel, self.instance_name)


@contextmanager
def upload(
    channel: grpc.Channel,
    instance: Optional[str] = None,
    u_uid: Optional[str] = None,
    retries: int = 0,
    max_backoff: int = 64,
    should_backoff: bool = True,
) -> Iterator[Uploader]:
    """Context manager generator for the :class:`Uploader` class."""
    uploader = Uploader(
        channel,
        instance=instance,
        u_uid=u_uid,
        retries=retries,
        max_backoff=max_backoff,
        should_backoff=should_backoff,
    )
    try:
        yield uploader
    finally:
        uploader.close()



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod
from contextvars import ContextVar
from typing import Any, Dict, List, Mapping, Optional, Set, Tuple, Type, Union, cast

import grpc
import jwt

from buildgrid._protos.buildgrid.v2.identity_pb2 import ClientIdentity
from buildgrid.server.auth.config import InstanceAuthorizationConfig
from buildgrid.server.auth.enums import AuthMetadataAlgorithm
from buildgrid.server.auth.exceptions import (
    AuthError,
    ExpiredTokenError,
    InvalidAuthorizationHeaderError,
    InvalidTokenError,
    MissingTokenError,
    SigningKeyNotFoundError,
    UnboundedTokenError,
    UnexpectedTokenParsingError,
)
from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.settings import AUTH_CACHE_SIZE

LOGGER = buildgrid_logger(__name__)


AlgorithmType = Union[
    Type[jwt.algorithms.RSAAlgorithm], Type[jwt.algorithms.ECAlgorithm], Type[jwt.algorithms.HMACAlgorithm]
]

# Algorithm classes defined in: https://github.com/jpadilla/pyjwt/blob/master/jwt/algorithms.py
ALGORITHM_TO_PYJWT_CLASS: Dict[str, AlgorithmType] = {
    "RSA": jwt.algorithms.RSAAlgorithm,
    "EC": jwt.algorithms.ECAlgorithm,
    "oct": jwt.algorithms.HMACAlgorithm,
}


def _log_and_raise(request_name: str, exception: AuthError) -> str:
    LOGGER.info("Authorization error. Rejecting.", tags=dict(request_name=request_name, reason=str(exception)))
    raise exception


class JwtParser:
    def __init__(
        self,
        secret: Optional[str] = None,
        algorithm: AuthMetadataAlgorithm = AuthMetadataAlgorithm.UNSPECIFIED,
        jwks_urls: Optional[List[str]] = None,
        audiences: Optional[List[str]] = None,
        jwks_fetch_minutes: int = 60,
    ) -> None:
        self._check_jwt_support(algorithm)

        self._algorithm = algorithm
        self._audiences = audiences

        if (secret is None and jwks_urls is None) or (secret is not None and jwks_urls is not None):
            raise TypeError("Exactly one of `secret` or `jwks_url` must be set")
        self._secret = secret
        self._jwks_clients = [
            jwt.PyJWKClient(url, lifespan=60 * jwks_fetch_minutes, max_cached_keys=AUTH_CACHE_SIZE)
            for url in (jwks_urls or [])
        ]

    def _check_jwt_support(self, algorithm: AuthMetadataAlgorithm) -> None:
        """Ensures JWT and possible dependencies are available."""
        if algorithm == AuthMetadataAlgorithm.UNSPECIFIED:
            raise InvalidArgumentError("JWT authorization method requires an algorithm to be specified")

    def parse(self, token: str) -> Dict[str, Any]:
        payload: Optional[Dict[str, Any]] = None
        try:
            if self._secret is not None:
                payload = jwt.decode(
                    token,
                    self._secret,
                    algorithms=[self._algorithm.value.upper()],
                    audience=self._audiences,
                    options={"require": ["exp"], "verify_exp": True},
                )

            elif self._jwks_clients:
                # Find the signing_key in jkus
                signing_key: Optional[jwt.PyJWK] = None
                errors: List[Tuple[str, jwt.PyJWKClientError]] = []
                for jwks_client in self._jwks_clients:
                    try:
                        signing_key = jwks_client.get_signing_key_from_jwt(token)
                        break
                    except jwt.PyJWKClientError as e:
                        errors.append((jwks_client.uri, e))

                if signing_key is None:
                    error_msg = ", ".join(f"{uri}:{str(err)}" for uri, err in errors)
                    raise SigningKeyNotFoundError(error_msg)

                payload = jwt.decode(
                    token,
                    signing_key.key,
                    algorithms=[self._algorithm.value.upper()],
                    audience=self._audiences,
                    options={"require": ["exp"], "verify_exp": True},
                )

        except AuthError:
            raise

        except jwt.exceptions.ExpiredSignatureError as e:
            raise ExpiredTokenError() from e

        except jwt.exceptions.MissingRequiredClaimError as e:
            raise UnboundedTokenError("Missing required JWT claim, likely 'exp' was not set") from e

        except jwt.exceptions.InvalidTokenError as e:
            raise InvalidTokenError() from e

        except Exception as e:
            raise UnexpectedTokenParsingError() from e

        if payload is None:
            raise InvalidTokenError()

        return payload

    def identity_from_jwt_payload(self, payload: Dict[str, Any]) -> ClientIdentity:
        """
        Extract the relevant claims from the JWT
            "aud" -> workflow
            "sub" -> subject
            "act" -> actor
        If the "act" field is not set then the subject is considered the actor
        The audience for the identity is taken from the config if set
        If "aud" field is an array of strings then the first element is set as the audience'
        Args:
            payload: the decoded payload from the jwt
        Returns:
            A dictionary containing workflow, actor, subject
        """

        workflow = ""
        if audience_from_payload := payload.get("aud"):
            if isinstance(audience_from_payload, str):
                workflow = audience_from_payload
            elif isinstance(audience_from_payload, list):
                workflow = audience_from_payload[0]
        elif self._audiences is not None and len(self._audiences) > 0:
            workflow = self._audiences[0]

        actor = payload.get("act")
        subject = payload.get("sub")

        if not actor:
            actor = subject
        return ClientIdentity(
            actor=actor if isinstance(actor, str) else "",
            subject=subject if isinstance(subject, str) else "",
            workflow=workflow if isinstance(workflow, str) else "",
        )

    def identity_from_token(self, token: str) -> ClientIdentity:
        payload = self.parse(token)
        return self.identity_from_jwt_payload(payload)


class AuthManager(ABC):
    @abstractmethod
    def authorize(self, context: grpc.ServicerContext, instance_name: str, request_name: str) -> bool:
        """Determine whether or not a request is authorized.

        This method takes a ``ServicerContext`` for an incoming gRPC request,
        along with the name of the request, and the name of the instance that
        the request is intended for. Information about the identity of the
        requester is extracted from the context, for example a JWT token.

        This identity information is compared to the ACL configuration given
        to this class at construction time to determine authorization for the
        request.

        Args:
            context (ServicerContext): The context for the gRPC request to check
                the authz status of.

            instance_name (str): The name of the instance that the gRPC request
                will be interacting with. This is used for per-instance ACLs.

            request_name (str): The name of the request being authorized, for
                example `Execute`.

        Returns:
            bool: Whether the request is authorized.

        """


class JWTAuthManager(AuthManager):
    def __init__(
        self,
        secret: Optional[str] = None,
        algorithm: AuthMetadataAlgorithm = AuthMetadataAlgorithm.UNSPECIFIED,
        jwks_urls: Optional[List[str]] = None,
        audiences: Optional[List[str]] = None,
        jwks_fetch_minutes: int = 60,
        acls: Optional[Mapping[str, InstanceAuthorizationConfig]] = None,
        allow_unauthorized_instances: Optional[Set[str]] = None,
    ) -> None:
        """Initializes a new :class:`JWTAuthManager`.

        Args:
            secret (str): The secret or key to be used for validating request,
                depending on `method`. Defaults to ``None``.

            algorithm (AuthMetadataAlgorithm): The crytographic algorithm used
                to encode `secret`. Defaults to ``UNSPECIFIED``.

            jwks_urls (list[str]): The urls to fetch the JWKs. Either secret or
                this field must be specified if the authentication method is JWT.
                Defaults to ``None``.

            audiences (list[str]): The audience used to validate jwt tokens against.
                The tokens must have an audience field.

            jwks_fetch_minutes (int): The number of minutes to cache JWKs fetches
                for. Defaults to 60.

            acls (Mapping[str, InstanceAuthorizationConfig] | None): An optional
                map of instance name -> ACL config to use for per-instance
                authorization.

            allow_unauthorized_instances(Set[str] | None): List of instances that should
                be allowed to have unautheticated access

        Raises:
            InvalidArgumentError: If `algorithm` is not supported.

        """
        self._acls = acls
        self._allow_unauthorized_instances = allow_unauthorized_instances
        self._token_parser = JwtParser(secret, algorithm, jwks_urls, audiences, jwks_fetch_minutes)

    def _token_from_request_context(self, context: grpc.ServicerContext, request_name: str) -> str:
        try:
            bearer = cast(str, dict(context.invocation_metadata())["authorization"])

        except KeyError:
            # Reject requests not carrying a token
            _log_and_raise(request_name, MissingTokenError())

        # Reject requests with malformatted bearer
        if not bearer.startswith("Bearer "):
            _log_and_raise(request_name, InvalidAuthorizationHeaderError())

        return bearer[7:]

    def authorize(self, context: grpc.ServicerContext, instance_name: str, request_name: str) -> bool:
        # No need to authorize if unauthorized access is allowed for the instance
        if self._allow_unauthorized_instances and instance_name in self._allow_unauthorized_instances:
            return True
        try:
            token = self._token_from_request_context(context, request_name)
            identity_from_token = self._token_parser.identity_from_token(token)
            workflow = identity_from_token.workflow
            actor = identity_from_token.actor
            subject = identity_from_token.subject
            set_context_client_identity(identity_from_token)
        except NameError:
            LOGGER.error("JWT auth is enabled but PyJWT is not installed.")
            return False
        except AuthError as e:
            LOGGER.info(f"Error authorizing JWT token: {str(e)}")
            return False

        # If no ACL config was provided at all, don't do any more validation
        if self._acls is None:
            return True

        instance_acl_config = self._acls.get(instance_name)
        if instance_acl_config is not None:
            return instance_acl_config.is_authorized(request_name, actor=actor, subject=subject, workflow=workflow)

        # If there is an ACL, but no config for this instance, deny all
        return False


class HeadersAuthManager(AuthManager):
    def __init__(
        self,
        acls: Optional[Mapping[str, InstanceAuthorizationConfig]] = None,
        allow_unauthorized_instances: Optional[Set[str]] = None,
    ) -> None:
        """Initializes a new :class:`HeadersAuthManager`.

        Args:
            acls (Mapping[str, InstanceAuthorizationConfig] | None): An optional
                map of instance name -> ACL config to use for per-instance
                authorization.

            allow_unauthorized_instances(Set[str] | None): List of instances that should
                be allowed to have unautheticated access

        """
        self._acls = acls
        self._allow_unauthorized_instances = allow_unauthorized_instances

    def authorize(self, context: grpc.ServicerContext, instance_name: str, request_name: str) -> bool:
        # No need to authorize if unauthorized access is allowed for the instance
        if self._allow_unauthorized_instances and instance_name in self._allow_unauthorized_instances:
            return True
        metadata_dict = dict(context.invocation_metadata())
        actor = str(metadata_dict.get("x-request-actor"))
        subject = str(metadata_dict.get("x-request-subject"))
        workflow = str(metadata_dict.get("x-request-workflow"))
        set_context_client_identity(ClientIdentity(actor=actor, subject=subject, workflow=workflow))
        # If no ACL config was provided at all, don't do any more validation
        if self._acls is None:
            return True

        instance_acl_config = self._acls.get(instance_name)
        if instance_acl_config is not None:
            return instance_acl_config.is_authorized(request_name, actor=actor, subject=subject, workflow=workflow)

        # If there is an ACL, but no config for this instance, deny all
        return False


# TODO: Once https://github.com/grpc/grpc/issues/33071 is resolved this AuthContext can be
# replaced with a gRPC interceptor stores the AuthManager in a request-local ContextVar.
AuthContext: "ContextVar[Optional[AuthManager]]" = ContextVar("AuthManager", default=None)


def set_auth_manager(manager: Optional[AuthManager]) -> None:
    AuthContext.set(manager)


def get_auth_manager() -> Optional[AuthManager]:
    return AuthContext.get()


def authorize_request(request_context: grpc.ServicerContext, instance_name: str, request_name: str) -> None:
    manager = get_auth_manager()

    # If no auth is configured, don't do authz
    if manager is None:
        return

    if manager.authorize(request_context, instance_name, request_name):
        return

    LOGGER.info(
        "Authentication failed for request.", tags=dict(request_name=request_name, peer=request_context.peer())
    )
    # No need to yield here since calling `abort` raises an Exception
    request_context.abort(grpc.StatusCode.UNAUTHENTICATED, "No valid authorization or authentication provided")


ContextClientIdentity: "ContextVar[Optional[ClientIdentity]]" = ContextVar("ClientIdentity", default=None)


def set_context_client_identity(clientIdentity: ClientIdentity) -> None:
    ContextClientIdentity.set(clientIdentity)


def get_context_client_identity() -> Optional[ClientIdentity]:
    return ContextClientIdentity.get()



# Copyright (C) 2019 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from typing import Any

import click
from google.protobuf import json_format

from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2
from buildgrid.server.client.actioncache import query
from buildgrid.server.client.cas import download
from buildgrid.server.client.channel import setup_channel
from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.utils.digests import create_digest, parse_digest

from ..cli import pass_context


@click.group(name="action-cache", short_help="Query and update the action cache service.")
@click.option(
    "--remote",
    type=click.STRING,
    default="http://localhost:50051",
    show_default=True,
    help="Remote execution server's URL (port defaults to 50051 if no specified).",
)
@click.option(
    "--auth-token",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Authorization token for the remote.",
)
@click.option(
    "--client-key",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Private client key for TLS (PEM-encoded).",
)
@click.option(
    "--client-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public client certificate for TLS (PEM-encoded).",
)
@click.option(
    "--server-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public server certificate for TLS (PEM-encoded)",
)
@click.option(
    "--instance-name", type=click.STRING, default=None, show_default=True, help="Targeted farm instance name."
)
@click.option("--action-id", type=str, help="Action ID.")
@click.option("--invocation-id", type=str, help="Tool invocation ID.")
@click.option("--correlation-id", type=str, help="Correlated invocation ID.")
@pass_context
def cli(
    context: Any,
    remote: str,
    instance_name: str,
    auth_token: str,
    client_key: str,
    client_cert: str,
    server_cert: str,
    action_id: str,
    invocation_id: str,
    correlation_id: str,
) -> None:
    """Entry-point for the ``bgd action-cache`` CLI command group."""
    try:
        context.channel, _ = setup_channel(
            remote,
            auth_token=auth_token,
            client_key=client_key,
            client_cert=client_cert,
            server_cert=server_cert,
            action_id=action_id,
            tool_invocation_id=invocation_id,
            correlated_invocations_id=correlation_id,
        )

    except InvalidArgumentError as e:
        click.echo(f"Error: {e}.", err=True)
        sys.exit(-1)

    context.instance_name = instance_name


@cli.command("get", short_help="Retrieves a cached action-result.")
@click.argument("action-digest-string", nargs=1, type=click.STRING, required=True)
@click.option("--json", is_flag=True, show_default=True, help="Print action result in JSON format.")
@pass_context
def get(context: Any, action_digest_string: str, json: bool) -> None:
    """Entry-point of the ``bgd action-cache get`` CLI command.

    Note:
        Digest strings are expected to be like: ``{hash}/{size_bytes}``.
    """
    action_digest = parse_digest(action_digest_string)
    if action_digest is None:
        click.echo(f"Error: Invalid digest string '{action_digest_string}'.", err=True)
        sys.exit(-1)

    # Simply hit the action cache with the given action digest:
    with query(context.channel) as action_cache:
        try:
            action_result = action_cache.get(context.instance_name, action_digest)
        except Exception as e:
            click.echo(f"Error: Fetching from the action cache: {e}", err=True)
            sys.exit(-1)

    if action_result is not None:
        if not json:
            action_result_digest = create_digest(action_result.SerializeToString())

            click.echo(
                f"Hit: {action_digest.hash[:8]}/{action_digest.size_bytes}: "
                f"Result cached with digest=[{action_result_digest.hash}/{action_result_digest.size_bytes}]"
            )

            # TODO: Print ActionResult details?

        else:
            click.echo(json_format.MessageToJson(action_result))

    else:
        click.echo(
            f"Miss: {action_digest.hash[:8]}/{action_digest.size_bytes}: No associated result found in cache..."
        )


@cli.command("update", short_help="Maps an action to a given action-result.")
@click.argument("action-digest-string", nargs=1, type=click.STRING, required=True)
@click.argument("action-result-digest-string", nargs=1, type=click.STRING, required=True)
@pass_context
def update(context: Any, action_digest_string: str, action_result_digest_string: str) -> None:
    """Entry-point of the ``bgd action-cache update`` CLI command.

    Note:
        Digest strings are expected to be like: ``{hash}/{size_bytes}``.
    """
    action_digest = parse_digest(action_digest_string)
    if action_digest is None:
        click.echo(f"Error: Invalid digest string '{action_digest_string}'.", err=True)
        sys.exit(-1)

    action_result_digest = parse_digest(action_result_digest_string)
    if action_result_digest is None:
        click.echo(f"Error: Invalid digest string '{action_result_digest_string}'.", err=True)
        sys.exit(-1)

    # We have to download the ActionResult message from CAS first...
    with download(context.channel, instance=context.instance_name) as downloader:
        try:
            action_result = downloader.get_message(action_result_digest, remote_execution_pb2.ActionResult())
        except Exception as e:
            click.echo(f"Error: Fetching ActionResult from CAS: {e}", err=True)
            sys.exit(-1)

        # And only then we can update the action cache for the given digest:
        with query(context.channel) as action_cache:
            try:
                action_result_update = action_cache.update(context.instance_name, action_digest, action_result)
            except Exception as e:
                click.echo(f"Error: Uploading to ActionCache: {e}", err=True)
                sys.exit(-1)

            if action_result_update is None:
                click.echo(
                    "Error: Failed updating cache result for action="
                    f"[{action_digest.hash}/{action_digest.size_bytes}].",
                    err=True,
                )
                sys.exit(-1)



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A storage provider that uses redis to maintain existence and expiry metadata
for a storage.
"""


import time
from datetime import datetime, timedelta
from typing import IO, Dict, Iterator, List, Optional, Set, Tuple

import redis

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid._protos.google.rpc import code_pb2
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.cas.storage.index.index_abc import IndexABC
from buildgrid.server.cas.storage.storage_abc import StorageABC
from buildgrid.server.decorators import timed
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_gauge_metric
from buildgrid.server.redis.provider import RedisProvider
from buildgrid.server.utils.digests import validate_digest_data

LOGGER = buildgrid_logger(__name__)


class RedisIndex(IndexABC):
    TYPE = "RedisIndex"

    def __init__(self, redis: RedisProvider, storage: StorageABC) -> None:
        self._redis = redis
        self._storage = storage
        # TODO: implement redis notification based cleanup, make this configurable, and lower the default
        self._ttl = timedelta(days=365)

        # Keep track of the last returned scan cursor
        # to not start at the beginning for each call to `delete_n_bytes`
        self._delete_n_bytes_cursor = 0

    def start(self) -> None:
        self._storage.start()

    def stop(self) -> None:
        self._storage.stop()

    def _construct_key(self, digest: Digest) -> str:
        """Helper to get the redis key name for a particular digest"""
        # The tag prefix serves to distinguish between our keys and
        # actual blobs if the same redis is used for both index and storage
        return "A:" + digest.hash + "_" + str(digest.size_bytes)

    def _deconstruct_key(self, keystr: str) -> Optional[Digest]:
        """Helper to attempt to recover a Digest from a redis key"""

        try:
            tag, rest = keystr.split(":", 1)
            if tag != "A":
                return None
            hash, size_bytes = rest.rsplit("_", 1)
            return Digest(hash=hash, size_bytes=int(size_bytes))
        except ValueError:
            return None

    @timed(METRIC.STORAGE.STAT_DURATION, type=TYPE)
    def has_blob(self, digest: Digest) -> bool:
        # Redis is authoritative for existence, no need to check storage.
        return bool(self._redis.execute_ro(lambda r: r.exists(self._construct_key(digest))))

    @timed(METRIC.STORAGE.READ_DURATION, type=TYPE)
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        if blob := self._storage.get_blob(digest):
            return blob

        deleted_index_digests = self._bulk_delete_from_index([digest])
        for digest in deleted_index_digests:
            LOGGER.warning("Blob was indexed but not in storage. Deleted from the index.", tags=dict(digest=digest))

        return None

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    def delete_blob(self, digest: Digest) -> None:
        # If the initial delete doesn't delete anything due to the key not existing
        # don't do anything else
        if self._redis.execute_rw(lambda r: r.delete(self._construct_key(digest))):
            self._storage.delete_blob(digest)

            # If we race with a blob being re-added we might have just deleted the
            # storage out from under it. We don't want the index to end up with
            # keys for things that are not present in storage since we consider
            # the index authoritative for existance. So we delete the keys again
            # after deleting from storage, this way if they do get out of sync it
            # will be in the direction of leaking objects in storage that the
            # index doesn't know about.
            def delete_from_index(r: "redis.Redis[bytes]") -> None:
                pipe = r.pipeline()
                pipe.delete(self._construct_key(digest))
                pipe.decrby("total_size", digest.size_bytes)
                pipe.execute()

            self._redis.execute_rw(delete_from_index)

    @timed(METRIC.STORAGE.WRITE_DURATION, type=TYPE)
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        self._storage.commit_write(digest, write_session)

        def set_ttl(r: "redis.Redis[bytes]") -> None:
            key = self._construct_key(digest)
            # Only increment total_size if this key is new
            # Use a dummy value of 1. We only care about existence and expiry
            if r.set(key, 1, ex=self._ttl, nx=True) is not None:
                r.incrby("total_size", digest.size_bytes)

        self._redis.execute_rw(set_ttl)

    def _bulk_delete_from_index(self, digests: List[Digest]) -> List[Digest]:
        def delete_from_index(r: "redis.Redis[bytes]") -> List[Digest]:
            pipe = r.pipeline()
            bytes_deleted = 0
            for digest in digests:
                pipe.delete(self._construct_key(digest))
            results = pipe.execute()
            # Go through the delete calls and only decrement total_size for the keys
            # which were actually removed
            successful_deletes = []
            for result, digest in zip(results, digests):
                if result:
                    bytes_deleted += digest.size_bytes
                    successful_deletes.append(digest)
            r.decrby("total_size", bytes_deleted)
            return successful_deletes

        successful_deletes = self._redis.execute_rw(delete_from_index)
        return successful_deletes

    @timed(METRIC.STORAGE.BULK_DELETE_DURATION, type=TYPE)
    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        # Delete from the index and then delete from the backing storage.
        successful_deletes = self._bulk_delete_from_index(digests)
        failed_deletes = self._storage.bulk_delete(successful_deletes)
        return failed_deletes

    @timed(METRIC.STORAGE.BULK_STAT_DURATION, type=TYPE)
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        # We hit the RW node for every FMB call to extend all the TTLs.
        # This could try to take advantage of RO replicas by only hitting the
        # RW node for blobs that do not have enough TTL left, if any.
        # We currently rely on the always-updated TTL to determine if a blob
        # should be protected in mark_n_bytes_as_deleted. If we allow some
        # slop before updating the RW node here we need to account for it
        # there too.
        def extend_ttls(r: "redis.Redis[bytes]") -> List[int]:
            pipe = r.pipeline(transaction=False)
            for digest in digests:
                pipe.expire(name=self._construct_key(digest), time=self._ttl)
            return pipe.execute()

        extend_results = self._redis.execute_rw(extend_ttls)

        return [digest for digest, result in zip(digests, extend_results) if result != 1]

    @timed(METRIC.STORAGE.BULK_WRITE_DURATION, type=TYPE)
    def bulk_update_blobs(self, blobs: List[Tuple[Digest, bytes]]) -> List[Status]:
        result_map: Dict[str, Status] = {}
        missing_blob_pairs: List[Tuple[Digest, bytes]] = []
        missing_blobs = self.missing_blobs([digest for digest, _ in blobs])
        for digest, blob in blobs:
            if digest not in missing_blobs:
                if validate_digest_data(digest, blob):
                    result_map[digest.hash] = Status(code=code_pb2.OK)
                else:
                    result_map[digest.hash] = Status(code=code_pb2.INVALID_ARGUMENT, message="Data doesn't match hash")
            else:
                missing_blob_pairs.append((digest, blob))
        results = self._storage.bulk_update_blobs(missing_blob_pairs)

        def set_ttls(r: "redis.Redis[bytes]") -> None:
            pipe = r.pipeline()
            bytes_added = 0
            for digest, result in zip(missing_blobs, results):
                result_map[digest.hash] = result
                if result.code == code_pb2.OK:
                    key = self._construct_key(digest)
                    # Use a dummy value of 1. We only care about existence and expiry
                    pipe.set(key, 1, ex=self._ttl, nx=True)
            redis_results = pipe.execute()
            # only update total_size for brand new keys
            for result, digest in zip(redis_results, missing_blobs):
                if result is not None:
                    bytes_added += digest.size_bytes
            r.incrby("total_size", bytes_added)

        self._redis.execute_rw(set_ttls)
        return [result_map[digest.hash] for digest, _ in blobs]

    @timed(METRIC.STORAGE.BULK_READ_DURATION, type=TYPE)
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        fetched_digests = self._storage.bulk_read_blobs(digests)

        fetched_digest_hashes = set(digest_hash for (digest_hash, _) in fetched_digests.items())
        digests_not_in_storage: List[Digest] = []
        for expected_digest in digests:
            if expected_digest.hash not in fetched_digest_hashes:
                digests_not_in_storage.append(expected_digest)

        if digests_not_in_storage:
            deleted_index_digests = self._bulk_delete_from_index(digests_not_in_storage)
            for digest in deleted_index_digests:
                LOGGER.warning(
                    "Blob was indexed but not in storage. Deleted from the index.", tags=dict(digest=digest)
                )

        return fetched_digests

    def least_recent_digests(self) -> Iterator[Digest]:
        """Generator to iterate through the digests in LRU order"""
        # This is not a LRU index, this method is used only from tests.
        raise NotImplementedError()

    def get_total_size(self) -> int:
        """
        Return the sum of the size of all blobs within the index
        """

        # The total_size represents what we have stored in the underlying
        # storage. However, if some redis notifications for expiring keys
        # are missed we won't actually have keys to account for all the size.
        # The expectation is that a "janitor" process will locate orphaned
        # blobs in storage eventually and when it does so it will call our
        # delete_blob which will finally decrby the total_size.
        total_size = self._redis.execute_ro(lambda r: r.get("total_size"))
        if total_size:
            return int(total_size)
        else:
            return 0

    def delete_n_bytes(
        self, n_bytes: int, dry_run: bool = False, protect_blobs_after: Optional[datetime] = None
    ) -> int:
        """
        Iterate through the Redis Index using 'SCAN' and delete any entries older than
        'protect_blobs_after'. The ordering of the deletes is undefined and can't be assumed
        to be LRU.
        """
        now = datetime.utcnow()

        if protect_blobs_after:
            threshold_time = protect_blobs_after
        else:
            threshold_time = now

        seen: Set[str] = set()
        bytes_deleted = 0

        while n_bytes > 0:
            # Used for metric publishing
            delete_start_time = time.time()

            # Maybe count should be configurable or somehow self-tuning
            # based on how many deletable keys we're actually getting
            # back per-request.
            # We could also choose random prefixes for the scan so that
            # multiple cleanup process are less likely to contend
            rawkeys: List[bytes]
            previous_cursor = self._delete_n_bytes_cursor
            self._delete_n_bytes_cursor, rawkeys = self._redis.execute_ro(
                lambda r: r.scan(match="A:*", cursor=self._delete_n_bytes_cursor, count=1000)
            )
            keys = [key.decode() for key in rawkeys]

            def get_ttls(r: "redis.Redis[bytes]") -> List[bytes]:
                pipe = r.pipeline(transaction=False)
                for key in keys:
                    pipe.ttl(key)
                return pipe.execute()

            raw_ttls = self._redis.execute_ro(get_ttls)
            ttls = [int(x) for x in raw_ttls]

            LOGGER.debug("Scan returned.", tags=dict(key_count=len(ttls)))
            digests_to_delete: List[Digest] = []
            failed_deletes: List[str] = []
            for key, ttl in zip(keys, ttls):
                # Since FMB sets the ttl to self._ttl on every call we can
                # use the time remaining to figure out when the last FMB
                # call for that blob was.
                blob_time = now - (self._ttl - timedelta(seconds=ttl))
                if n_bytes <= 0:
                    # Reset scan cursor to previous value to not skip
                    # the digests we didn't get to
                    self._delete_n_bytes_cursor = previous_cursor
                    break

                digest = self._deconstruct_key(key)
                if digest and (blob_time <= threshold_time) and (digest.hash not in seen):
                    n_bytes -= digest.size_bytes
                    digests_to_delete.append(digest)

            if digests_to_delete:
                if dry_run:
                    LOGGER.debug("Detected deletable digests.", tags=dict(digest_count=len(digests_to_delete)))
                    for digest in digests_to_delete:
                        if digest not in failed_deletes:
                            bytes_deleted += digest.size_bytes
                else:
                    LOGGER.debug("Deleting digests.", tags=dict(digest_count=len(digests_to_delete)))
                    failed_deletes = self.bulk_delete(digests_to_delete)
                    blobs_deleted = 0
                    for digest in digests_to_delete:
                        if digest not in failed_deletes:
                            blobs_deleted += 1
                            bytes_deleted += digest.size_bytes

                    batch_duration = time.time() - delete_start_time
                    blobs_deleted_per_second = blobs_deleted / batch_duration
                    publish_gauge_metric(METRIC.CLEANUP.BLOBS_DELETED_PER_SECOND, blobs_deleted_per_second)

            if self._delete_n_bytes_cursor == 0:  # scan finished
                LOGGER.debug("Cursor exhausted.")
                break

        return bytes_deleted



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2020 The Bazel Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.duration_pb2
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.timestamp_pb2
import buildgrid._protos.google.rpc.status_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Qualifier(google.protobuf.message.Message):
    """The Remote Asset API provides a mapping from a URI and Qualifiers to
    Digests.

    Multiple URIs may be used to refer to the same content.  For example, the
    same tarball may exist at multiple mirrors and thus be retrievable from
    multiple URLs.  When URLs are used, these should refer to actual content as
    Fetch service implementations may choose to fetch the content directly
    from the origin.  For example, the HEAD of a git repository's active branch
    can be referred to as:

        uri: https://github.com/bazelbuild/remote-apis.git

    URNs may be used to strongly identify content, for instance by using the
    uuid namespace identifier: urn:uuid:f81d4fae-7dec-11d0-a765-00a0c91e6bf6.
    This is most applicable to named content that is Push'd, where the URN
    serves as an agreed-upon key, but carries no other inherent meaning.

    Service implementations may choose to support only URLs, only URNs for
    Push'd content, only other URIs for which the server and client agree upon
    semantics of, or any mixture of the above.

    Qualifiers are used to disambiguate or sub-select content that shares a URI.
    This may include specifying a particular commit or branch, in the case of
    URIs referencing a repository; they could also be used to specify a
    particular subdirectory of a repository or tarball. Qualifiers may also be
    used to ensure content matches what the client expects, even when there is
    no ambiguity to be had - for example, a qualifier specifying a checksum
    value.

    In cases where the semantics of the request are not immediately clear from
    the URL and/or qualifiers - e.g. dictated by URL scheme - it is recommended
    to use an additional qualifier to remove the ambiguity. The `resource_type`
    qualifier is recommended for this purpose.

    Qualifiers may be supplied in any order.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The "name" of the qualifier, for example "resource_type".
    No separation is made between 'standard' and 'nonstandard'
    qualifiers, in accordance with https://tools.ietf.org/html/rfc6648,
    however implementers *SHOULD* take care to avoid ambiguity.
    """
    value: builtins.str
    """The "value" of the qualifier. Semantics will be dictated by the name."""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        value: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name", "value", b"value"]) -> None: ...

global___Qualifier = Qualifier

@typing_extensions.final
class FetchBlobRequest(google.protobuf.message.Message):
    """A request message for
    [Fetch.FetchBlob][build.bazel.remote.asset.v1.Fetch.FetchBlob].
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INSTANCE_NAME_FIELD_NUMBER: builtins.int
    TIMEOUT_FIELD_NUMBER: builtins.int
    OLDEST_CONTENT_ACCEPTED_FIELD_NUMBER: builtins.int
    URIS_FIELD_NUMBER: builtins.int
    QUALIFIERS_FIELD_NUMBER: builtins.int
    instance_name: builtins.str
    """The instance of the execution system to operate against. A server may
    support multiple instances of the execution system (with their own workers,
    storage, caches, etc.). The server MAY require use of this field to select
    between them in an implementation-defined fashion, otherwise it can be
    omitted.
    """
    @property
    def timeout(self) -> google.protobuf.duration_pb2.Duration:
        """The timeout for the underlying fetch, if content needs to be retrieved from
        origin.

        If unset, the server *MAY* apply an implementation-defined timeout.

        If set, and the user-provided timeout exceeds the RPC deadline, the server
        *SHOULD* keep the fetch going after the RPC completes, to be made
        available for future Fetch calls. The server may also enforce (via clamping
        and/or an INVALID_ARGUMENT error) implementation-defined minimum and
        maximum timeout values.

        If this timeout is exceeded on an attempt to retrieve content from origin
        the client will receive DEADLINE_EXCEEDED in [FetchBlobResponse.status].
        """
    @property
    def oldest_content_accepted(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """The oldest content the client is willing to accept, as measured from the
        time it was Push'd or when the underlying retrieval from origin was
        started.
        Upon retries of Fetch requests that cannot be completed within a single
        RPC, clients *SHOULD* provide the same value for subsequent requests as the
        original, to simplify combining the request with the previous attempt.

        If unset, the client *SHOULD* accept content of any age.
        """
    @property
    def uris(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """The URI(s) of the content to fetch. These may be resources that the server
        can directly fetch from origin, in which case multiple URIs *SHOULD*
        represent the same content available at different locations (such as an
        origin and secondary mirrors). These may also be URIs for content known to
        the server through other mechanisms, e.g. pushed via the [Push][build.bazel.remote.asset.v1.Push]
        service.

        Clients *MUST* supply at least one URI. Servers *MAY* match any one of the
        supplied URIs.
        """
    @property
    def qualifiers(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Qualifier]:
        """Qualifiers sub-specifying the content to fetch - see comments on
        [Qualifier][build.bazel.remote.asset.v1.Qualifier].
        The same qualifiers apply to all URIs.

        Specified qualifier names *MUST* be unique.
        """
    def __init__(
        self,
        *,
        instance_name: builtins.str = ...,
        timeout: google.protobuf.duration_pb2.Duration | None = ...,
        oldest_content_accepted: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        uris: collections.abc.Iterable[builtins.str] | None = ...,
        qualifiers: collections.abc.Iterable[global___Qualifier] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["oldest_content_accepted", b"oldest_content_accepted", "timeout", b"timeout"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["instance_name", b"instance_name", "oldest_content_accepted", b"oldest_content_accepted", "qualifiers", b"qualifiers", "timeout", b"timeout", "uris", b"uris"]) -> None: ...

global___FetchBlobRequest = FetchBlobRequest

@typing_extensions.final
class FetchBlobResponse(google.protobuf.message.Message):
    """A response message for
    [Fetch.FetchBlob][build.bazel.remote.asset.v1.Fetch.FetchBlob].
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATUS_FIELD_NUMBER: builtins.int
    URI_FIELD_NUMBER: builtins.int
    QUALIFIERS_FIELD_NUMBER: builtins.int
    EXPIRES_AT_FIELD_NUMBER: builtins.int
    BLOB_DIGEST_FIELD_NUMBER: builtins.int
    @property
    def status(self) -> buildgrid._protos.google.rpc.status_pb2.Status:
        """If the status has a code other than `OK`, it indicates that the operation
        was unable to be completed for reasons outside the servers' control.
        The possible fetch errors include:
        * `DEADLINE_EXCEEDED`: The operation could not be completed within the
          specified timeout.
        * `NOT_FOUND`: The requested asset was not found at the specified location.
        * `PERMISSION_DENIED`: The request was rejected by a remote server, or
          requested an asset from a disallowed origin.
        * `ABORTED`: The operation could not be completed, typically due to a
          failed consistency check.
        * `RESOURCE_EXHAUSTED`: There is insufficient quota of some resource to
          perform the requested operation. The client may retry after a delay.
        """
    uri: builtins.str
    """The uri from the request that resulted in a successful retrieval, or from
    which the error indicated in `status` was obtained.
    """
    @property
    def qualifiers(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Qualifier]:
        """Any qualifiers known to the server and of interest to clients."""
    @property
    def expires_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """A minimum timestamp the content is expected to be available through.
        Servers *MAY* omit this field, if not known with confidence.
        """
    @property
    def blob_digest(self) -> buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest:
        """The result of the fetch, if the status had code `OK`.
        The digest of the file's contents, available for download through the CAS.
        """
    def __init__(
        self,
        *,
        status: buildgrid._protos.google.rpc.status_pb2.Status | None = ...,
        uri: builtins.str = ...,
        qualifiers: collections.abc.Iterable[global___Qualifier] | None = ...,
        expires_at: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        blob_digest: buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["blob_digest", b"blob_digest", "expires_at", b"expires_at", "status", b"status"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["blob_digest", b"blob_digest", "expires_at", b"expires_at", "qualifiers", b"qualifiers", "status", b"status", "uri", b"uri"]) -> None: ...

global___FetchBlobResponse = FetchBlobResponse

@typing_extensions.final
class FetchDirectoryRequest(google.protobuf.message.Message):
    """A request message for
    [Fetch.FetchDirectory][build.bazel.remote.asset.v1.Fetch.FetchDirectory].
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INSTANCE_NAME_FIELD_NUMBER: builtins.int
    TIMEOUT_FIELD_NUMBER: builtins.int
    OLDEST_CONTENT_ACCEPTED_FIELD_NUMBER: builtins.int
    URIS_FIELD_NUMBER: builtins.int
    QUALIFIERS_FIELD_NUMBER: builtins.int
    instance_name: builtins.str
    """The instance of the execution system to operate against. A server may
    support multiple instances of the execution system (with their own workers,
    storage, caches, etc.). The server MAY require use of this field to select
    between them in an implementation-defined fashion, otherwise it can be
    omitted.
    """
    @property
    def timeout(self) -> google.protobuf.duration_pb2.Duration:
        """The timeout for the underlying fetch, if content needs to be retrieved from
        origin. This value is allowed to exceed the RPC deadline, in which case the
        server *SHOULD* keep the fetch going after the RPC completes, to be made
        available for future Fetch calls.

        If this timeout is exceeded on an attempt to retrieve content from origin
        the client will receive DEADLINE_EXCEEDED in [FetchDirectoryResponse.status].
        """
    @property
    def oldest_content_accepted(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """The oldest content the client is willing to accept, as measured from the
        time it was Push'd or when the underlying retrieval from origin was
        started.
        Upon retries of Fetch requests that cannot be completed within a single
        RPC, clients *SHOULD* provide the same value for subsequent requests as the
        original, to simplify combining the request with the previous attempt.

        If unset, the client *SHOULD* accept content of any age.
        """
    @property
    def uris(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """The URI(s) of the content to fetch. These may be resources that the server
        can directly fetch from origin, in which case multiple URIs *SHOULD*
        represent the same content available at different locations (such as an
        origin and secondary mirrors). These may also be URIs for content known to
        the server through other mechanisms, e.g. pushed via the [Push][build.bazel.remote.asset.v1.Push]
        service.

        Clients *MUST* supply at least one URI. Servers *MAY* match any one of the
        supplied URIs.
        """
    @property
    def qualifiers(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Qualifier]:
        """Qualifiers sub-specifying the content to fetch - see comments on
        [Qualifier][build.bazel.remote.asset.v1.Qualifier].
        The same qualifiers apply to all URIs.

        Specified qualifier names *MUST* be unique.
        """
    def __init__(
        self,
        *,
        instance_name: builtins.str = ...,
        timeout: google.protobuf.duration_pb2.Duration | None = ...,
        oldest_content_accepted: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        uris: collections.abc.Iterable[builtins.str] | None = ...,
        qualifiers: collections.abc.Iterable[global___Qualifier] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["oldest_content_accepted", b"oldest_content_accepted", "timeout", b"timeout"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["instance_name", b"instance_name", "oldest_content_accepted", b"oldest_content_accepted", "qualifiers", b"qualifiers", "timeout", b"timeout", "uris", b"uris"]) -> None: ...

global___FetchDirectoryRequest = FetchDirectoryRequest

@typing_extensions.final
class FetchDirectoryResponse(google.protobuf.message.Message):
    """A response message for
    [Fetch.FetchDirectory][build.bazel.remote.asset.v1.Fetch.FetchDirectory].
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATUS_FIELD_NUMBER: builtins.int
    URI_FIELD_NUMBER: builtins.int
    QUALIFIERS_FIELD_NUMBER: builtins.int
    EXPIRES_AT_FIELD_NUMBER: builtins.int
    ROOT_DIRECTORY_DIGEST_FIELD_NUMBER: builtins.int
    @property
    def status(self) -> buildgrid._protos.google.rpc.status_pb2.Status:
        """If the status has a code other than `OK`, it indicates that the operation
        was unable to be completed for reasons outside the servers' control.
        The possible fetch errors include:
        * `DEADLINE_EXCEEDED`: The operation could not be completed within the
          specified timeout.
        * `NOT_FOUND`: The requested asset was not found at the specified location.
        * `PERMISSION_DENIED`: The request was rejected by a remote server, or
          requested an asset from a disallowed origin.
        * `ABORTED`: The operation could not be completed, typically due to a
          failed consistency check.
        * `RESOURCE_EXHAUSTED`: There is insufficient quota of some resource to
          perform the requested operation. The client may retry after a delay.
        """
    uri: builtins.str
    """The uri from the request that resulted in a successful retrieval, or from
    which the error indicated in `status` was obtained.
    """
    @property
    def qualifiers(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Qualifier]:
        """Any qualifiers known to the server and of interest to clients."""
    @property
    def expires_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """A minimum timestamp the content is expected to be available through.
        Servers *MAY* omit this field, if not known with confidence.
        """
    @property
    def root_directory_digest(self) -> buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest:
        """The result of the fetch, if the status had code `OK`.
        the root digest of a directory tree, suitable for fetching via
        [ContentAddressableStorage.GetTree].
        """
    def __init__(
        self,
        *,
        status: buildgrid._protos.google.rpc.status_pb2.Status | None = ...,
        uri: builtins.str = ...,
        qualifiers: collections.abc.Iterable[global___Qualifier] | None = ...,
        expires_at: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        root_directory_digest: buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["expires_at", b"expires_at", "root_directory_digest", b"root_directory_digest", "status", b"status"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["expires_at", b"expires_at", "qualifiers", b"qualifiers", "root_directory_digest", b"root_directory_digest", "status", b"status", "uri", b"uri"]) -> None: ...

global___FetchDirectoryResponse = FetchDirectoryResponse

@typing_extensions.final
class PushBlobRequest(google.protobuf.message.Message):
    """A request message for
    [Push.PushBlob][build.bazel.remote.asset.v1.Push.PushBlob].
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INSTANCE_NAME_FIELD_NUMBER: builtins.int
    URIS_FIELD_NUMBER: builtins.int
    QUALIFIERS_FIELD_NUMBER: builtins.int
    EXPIRE_AT_FIELD_NUMBER: builtins.int
    BLOB_DIGEST_FIELD_NUMBER: builtins.int
    REFERENCES_BLOBS_FIELD_NUMBER: builtins.int
    REFERENCES_DIRECTORIES_FIELD_NUMBER: builtins.int
    instance_name: builtins.str
    """The instance of the execution system to operate against. A server may
    support multiple instances of the execution system (with their own workers,
    storage, caches, etc.). The server MAY require use of this field to select
    between them in an implementation-defined fashion, otherwise it can be
    omitted.
    """
    @property
    def uris(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """The URI(s) of the content to associate. If multiple URIs are specified, the
        pushed content will be available to fetch by specifying any of them.
        """
    @property
    def qualifiers(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Qualifier]:
        """Qualifiers sub-specifying the content that is being pushed - see comments
        on [Qualifier][build.bazel.remote.asset.v1.Qualifier].
        The same qualifiers apply to all URIs.
        """
    @property
    def expire_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """A time after which this content should stop being returned via [FetchBlob][build.bazel.remote.asset.v1.Fetch.FetchBlob].
        Servers *MAY* expire content early, e.g. due to storage pressure.
        """
    @property
    def blob_digest(self) -> buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest:
        """The blob to associate."""
    @property
    def references_blobs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest]:
        """Referenced blobs or directories that need to not expire before expiration
        of this association, in addition to `blob_digest` itself.
        These fields are hints - clients *MAY* omit them, and servers *SHOULD*
        respect them, at the risk of increased incidents of Fetch responses
        indirectly referencing unavailable blobs.
        """
    @property
    def references_directories(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest]: ...
    def __init__(
        self,
        *,
        instance_name: builtins.str = ...,
        uris: collections.abc.Iterable[builtins.str] | None = ...,
        qualifiers: collections.abc.Iterable[global___Qualifier] | None = ...,
        expire_at: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        blob_digest: buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest | None = ...,
        references_blobs: collections.abc.Iterable[buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest] | None = ...,
        references_directories: collections.abc.Iterable[buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["blob_digest", b"blob_digest", "expire_at", b"expire_at"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["blob_digest", b"blob_digest", "expire_at", b"expire_at", "instance_name", b"instance_name", "qualifiers", b"qualifiers", "references_blobs", b"references_blobs", "references_directories", b"references_directories", "uris", b"uris"]) -> None: ...

global___PushBlobRequest = PushBlobRequest

@typing_extensions.final
class PushBlobResponse(google.protobuf.message.Message):
    """A response message for
    [Push.PushBlob][build.bazel.remote.asset.v1.Push.PushBlob].
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___PushBlobResponse = PushBlobResponse

@typing_extensions.final
class PushDirectoryRequest(google.protobuf.message.Message):
    """A request message for
    [Push.PushDirectory][build.bazel.remote.asset.v1.Push.PushDirectory].
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INSTANCE_NAME_FIELD_NUMBER: builtins.int
    URIS_FIELD_NUMBER: builtins.int
    QUALIFIERS_FIELD_NUMBER: builtins.int
    EXPIRE_AT_FIELD_NUMBER: builtins.int
    ROOT_DIRECTORY_DIGEST_FIELD_NUMBER: builtins.int
    REFERENCES_BLOBS_FIELD_NUMBER: builtins.int
    REFERENCES_DIRECTORIES_FIELD_NUMBER: builtins.int
    instance_name: builtins.str
    """The instance of the execution system to operate against. A server may
    support multiple instances of the execution system (with their own workers,
    storage, caches, etc.). The server MAY require use of this field to select
    between them in an implementation-defined fashion, otherwise it can be
    omitted.
    """
    @property
    def uris(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """The URI(s) of the content to associate. If multiple URIs are specified, the
        pushed content will be available to fetch by specifying any of them.
        """
    @property
    def qualifiers(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Qualifier]:
        """Qualifiers sub-specifying the content that is being pushed - see comments
        on [Qualifier][build.bazel.remote.asset.v1.Qualifier].
        The same qualifiers apply to all URIs.
        """
    @property
    def expire_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """A time after which this content should stop being returned via
        [FetchDirectory][build.bazel.remote.asset.v1.Fetch.FetchDirectory].
        Servers *MAY* expire content early, e.g. due to storage pressure.
        """
    @property
    def root_directory_digest(self) -> buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest:
        """Directory to associate"""
    @property
    def references_blobs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest]:
        """Referenced blobs or directories that need to not expire before expiration
        of this association, in addition to `root_directory_digest` itself.
        These fields are hints - clients *MAY* omit them, and servers *SHOULD*
        respect them, at the risk of increased incidents of Fetch responses
        indirectly referencing unavailable blobs.
        """
    @property
    def references_directories(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest]: ...
    def __init__(
        self,
        *,
        instance_name: builtins.str = ...,
        uris: collections.abc.Iterable[builtins.str] | None = ...,
        qualifiers: collections.abc.Iterable[global___Qualifier] | None = ...,
        expire_at: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        root_directory_digest: buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest | None = ...,
        references_blobs: collections.abc.Iterable[buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest] | None = ...,
        references_directories: collections.abc.Iterable[buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2.Digest] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["expire_at", b"expire_at", "root_directory_digest", b"root_directory_digest"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["expire_at", b"expire_at", "instance_name", b"instance_name", "qualifiers", b"qualifiers", "references_blobs", b"references_blobs", "references_directories", b"references_directories", "root_directory_digest", b"root_directory_digest", "uris", b"uris"]) -> None: ...

global___PushDirectoryRequest = PushDirectoryRequest

@typing_extensions.final
class PushDirectoryResponse(google.protobuf.message.Message):
    """A response message for
    [Push.PushDirectory][build.bazel.remote.asset.v1.Push.PushDirectory].
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___PushDirectoryResponse = PushDirectoryResponse



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Log Stream API"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class CreateLogStreamRequest(google.protobuf.message.Message):
    """Contains all information necessary to create a new LogStream resource."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PARENT_FIELD_NUMBER: builtins.int
    parent: builtins.str
    """Required. The parent resource of the created LogStream.
    The list of valid types of parent resources of LogStreams is up to the
    implementing server.
    Example: projects/123
    """
    def __init__(
        self,
        *,
        parent: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["parent", b"parent"]) -> None: ...

global___CreateLogStreamRequest = CreateLogStreamRequest

@typing_extensions.final
class LogStream(google.protobuf.message.Message):
    """A handle to a log (an ordered sequence of bytes)."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    WRITE_RESOURCE_NAME_FIELD_NUMBER: builtins.int
    name: builtins.str
    """Structured name of the resource in the format:
      {parent=**}/logstreams/{logstream_id}
      Example: projects/123/logStreams/456-def
    Attempting to call the Byte Stream API's `Write` RPC with a LogStream's
      `name` as the value for `ByteStream.Write.resource_name` is an error.
    """
    write_resource_name: builtins.str
    """Resource name to pass to `ByteStream.Write` in the format:
      {parent=**}/logStreams/{logstream_id}/{write_token}
      Example: projects/123/logStreams/456-def/789-ghi
    Attempting to call the Byte Stream API's `Read` RPC with a LogStream's
      `write_resource_name` as the value for `ByteStream.Write.resource_name`
      is an error.

    `write_resource_name` is separate from `name` to ensure that only the
    intended writers can write to a given LogStream. Writers must address write
    operations to the `write_resource_name`, not the `name`, and must have
    permission to write LogStreams. `write_resource_name` embeds a secret token
    and should be protected accordingly; a mishandled `write_resource_name` can
    result in unintended writers corrupting the LogStream. Therefore, the field
    should be excluded from calls to any calls which retrieve LogStream
    metadata (i.e.: `GetLogStream`).

    Bytes written to this resource must to be readable when `ByteStream.Read`
    is called with the `name` resource.
    Reading a write_resource_name must return an INVALID_ARGUMENT error.
    """
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        write_resource_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name", "write_resource_name", b"write_resource_name"]) -> None: ...

global___LogStream = LogStream



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from contextlib import contextmanager
from datetime import timedelta
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Generator, Iterator, List, Optional, Type, Union

from alembic import command
from alembic.config import Config
from importlib_resources import files
from sqlalchemy import create_engine, event, text
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session, scoped_session, sessionmaker
from sqlalchemy.pool import NullPool

from buildgrid.server.exceptions import DatabaseError, RetriableDatabaseError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.settings import (
    COOLDOWN_TIME_AFTER_POOL_DISPOSE_SECONDS,
    COOLDOWN_TIME_JITTER_BASE,
    MIN_TIME_BETWEEN_SQL_POOL_DISPOSE_MINUTES,
)

from .utils import (
    SQLPoolDisposeHelper,
    is_psycopg2_connection_string,
    is_sqlite_connection_string,
    is_sqlite_inmemory_connection_string,
)

LOGGER = buildgrid_logger(__name__)

# Each dialect has a limit on the number of bind parameters allowed. This
# matters because it determines how large we can allow our IN clauses to get.
#
# SQLite: 1000 https://www.sqlite.org/limits.html#max_variable_number
# PostgreSQL: 32767 (Int16.MAX_VALUE) https://www.postgresql.org/docs/9.4/protocol-message-formats.html
#
# We'll refer to this as the "inlimit" in the code. The inlimits are
# set to 75% of the bind parameter limit of the implementation.
DIALECT_INLIMIT_MAP = {"postgresql": 24000, "sqlite": 750}
DEFAULT_INLIMIT = 100


# NOTE: Obviously these type annotations are useless, but sadly they're what
# is in the upstream sqlalchemy2-stubs[0].
#
# Once we upgrade to SQLAlchemy 2.0 we can make these more useful, as that
# version of SQLAlchemy has sensible type annotations[1].
#
# [0]: https://github.com/sqlalchemy/sqlalchemy2-stubs/blob/main/sqlalchemy-stubs/pool/events.pyi#L9
# [1]: https://github.com/sqlalchemy/sqlalchemy/blob/main/lib/sqlalchemy/pool/events.py#L96-L100
def _sqlite_on_connect(conn: Any, record: Any) -> None:
    """SQLite ``PRAGMA`` statements to execute immediately upon connection.

    These statements configure the behaviour of the database, and are specific
    to SQLite.

    See https://www.sqlite.org/pragma.html for details.

    Args:
        conn (DBAPIConnection): The DBAPI connection that was just connected.
        record (_ConnectionRecord): The connection record which contains the
            DBAPI connection.

    """
    # Use journal_mode=WAL to allow read/write concurrency, as well as the
    # performance improvements it brings.
    conn.execute("PRAGMA journal_mode=WAL")
    conn.execute("PRAGMA synchronous=NORMAL")


class SqlProvider:
    """Class which provides an interface for interacting with an SQL database.

    This class is used to allow configuration of a per-process SQL connection
    pool, which can then be shared across multiple components of BuildGrid
    which require an SQL connection.

    Args:
        automigrate (bool): Whether or not to ensure the database is fully
            migrated when starting up. Defaults to ``False``, meaning the
            database is assumed to be at the up to date already.

        connection_string (str | None): The connection string to use when
            creating a database connection. If ``None`` then a temporary
            SQLite database will be created for the lifetime of this
            object.

        connection_timeout (int): The timeout to use when attempting to
            connect to the database, in seconds. Defaults to 5 seconds if
            unset.

        lock_timeout (int): The timeout to use when the connection
        holds a lock in the database. This is supported only if the database
        backend is PostgresQL.

        connect_args (Dict[str, Any] | None): Dictionary of DBAPI
            connection arguments to pass to the engine. See the
            SQLAlchemy `docs`_ for details.

        max_overflow (int | None): The number of connections to allow
            as "overflow" connections in the connection pool. This is
            the number of connections which will be able to be opened
            above the value of ``pool_size``.

        pool_pre_ping (bool | None): Whether or not to test connections
            for liveness on checkout from the connection pool.

        pool_recycle (int | None): The number of seconds after which to
            recycle database connections. If ``None`` (the default) then
            connections won't be recycled (the SQLAlchemy default value
            of -1 will be used).

        pool_size (int | None): The number of connections to keep open
            inside the engine's connection pool.

        pool_timeout (int | None): The number of seconds to wait when
            attempting to get a connection from the connection pool.

    Raises:
        ValueError: when ``connection_string`` specifies an in-memory SQLite
            database.

    .. _docs: https://docs.sqlalchemy.org/en/14/core/engines.html#use-the-connect-args-dictionary-parameter

    """

    def __init__(
        self,
        *,
        automigrate: bool = False,
        connection_string: Optional[str] = None,
        connection_timeout: int = 5,
        lock_timeout: int = 5,
        connect_args: Optional[Dict[Any, Any]] = None,
        max_overflow: Optional[int] = None,
        pool_pre_ping: Optional[bool] = None,
        pool_recycle: Optional[int] = None,
        pool_size: Optional[int] = None,
        pool_timeout: Optional[int] = None,
    ):
        """Initialize an SqlProvider."""
        self._database_tempfile = None
        # If we don't have a connection string, we'll make a tempfile to use
        # as an SQLite database. This tempfile needs to exist for the lifetime
        # of the SqlProvider.
        if not connection_string:
            self._database_tempfile = NamedTemporaryFile(prefix="bgd-", suffix=".db")
            LOGGER.warning(
                "No connection string specified for the SQL provider, will use SQLite with tempfile.",
                tags=dict(tempfile=self._database_tempfile.name),
            )
            automigrate = True  # since this is a temporary database, we always need to create it
            connection_string = f"sqlite:///{self._database_tempfile.name}"

        # Set up database connection
        self._session_factory = sessionmaker(future=True)
        self._scoped_session_factory = scoped_session(self._session_factory)

        self._engine = self._create_sqlalchemy_engine(
            connection_string,
            connection_timeout,
            lock_timeout=lock_timeout,
            connect_args=connect_args,
            max_overflow=max_overflow,
            pool_pre_ping=pool_pre_ping,
            pool_recycle=pool_recycle,
            pool_size=pool_size,
            pool_timeout=pool_timeout,
        )

        LOGGER.info("Created SQL provider.", tags=dict(automigrate=automigrate, connection=self._engine.url))

        if automigrate:
            self._create_or_migrate_db()

        self._sql_pool_dispose_helper = SQLPoolDisposeHelper(
            COOLDOWN_TIME_AFTER_POOL_DISPOSE_SECONDS,
            COOLDOWN_TIME_JITTER_BASE,
            MIN_TIME_BETWEEN_SQL_POOL_DISPOSE_MINUTES,
            self._engine,
        )

    def _create_sqlalchemy_engine(
        self,
        connection_string: str,
        connection_timeout: int,
        lock_timeout: int,
        *,
        connect_args: Optional[Dict[Any, Any]] = None,
        max_overflow: Optional[int] = None,
        pool_pre_ping: Optional[bool] = None,
        pool_recycle: Optional[int] = None,
        pool_size: Optional[int] = None,
        pool_timeout: Optional[int] = None,
    ) -> Engine:
        """Create the SQLAlchemy Engine.

        Args:
            connection_string: The connection string to use when
                creating the ``Engine``.

            connection_timeout: The timeout to use for database
                connections, in seconds. If set as 0, no timeout
                is applied.

            lock_timeout (int): The timeout to use when the connection
            holds a lock in the database. This is supported only if the database
            backend is PostgresQL.

            connect_args: Dictionary of DBAPI
                connection arguments to pass to the engine. See the
                SQLAlchemy `docs`_ for details.

            max_overflow: The number of connections to allow
                as "overflow" connections in the connection pool. This is
                the number of connections which will be able to be opened
                above the value of ``pool_size``.

            pool_pre_ping: Whether or not to test connections
                for liveness on checkout from the connection pool.

            pool_recycle: The number of seconds after which to
                recycle database connections. If ``None`` (the default) then
                connections won't be recycled (the SQLAlchemy default value
                of -1 will be used).

            pool_size: The number of connections to keep open
                inside the engine's connection pool.
                If set as zero, no connection pool is created
                and other pool_* parameters are ignored.

            pool_timeout: The number of seconds to wait when
                attempting to get a connection from the connection pool.

        Returns:
            A :class:`sqlalchemy.engine.Engine` set up to connect to the
                database defined by ``connection_string``.

        Raises:
            ValueError: when attempting to connect to an in-memory SQLite
                database.

        .. _docs: https://docs.sqlalchemy.org/en/14/core/engines.html#use-the-connect-args-dictionary-parameter

        """
        # Disallow sqlite in-memory because multi-threaded access to it is
        # complex and potentially problematic at best
        # ref: https://docs.sqlalchemy.org/en/14/dialects/sqlite.html#threading-pooling-behavior
        if is_sqlite_inmemory_connection_string(connection_string):
            raise ValueError(
                "Cannot use SQLite in-memory with BuildGrid "
                f"(connection_string=[{connection_string}]). Use a file or "
                "leave the connection_string empty for a tempfile."
            )

        extra_engine_args: Dict[str, Any] = {}
        if connect_args is not None:
            extra_engine_args["connect_args"] = connect_args
        else:
            extra_engine_args["connect_args"] = {}

        if connection_timeout > 0:
            if is_sqlite_connection_string(connection_string):
                extra_engine_args["connect_args"]["timeout"] = connection_timeout
            elif is_psycopg2_connection_string(connection_string):
                extra_engine_args["connect_args"]["connect_timeout"] = connection_timeout
        if lock_timeout > 0 and is_psycopg2_connection_string(connection_string):
            # Additional postgres specific timeouts
            # Additional libpg options
            # Note that those timeouts are in milliseconds (so *1000)
            # User might specifically set options... do not override in this case.
            extra_engine_args["connect_args"].setdefault("options", f"-c lock_timeout={lock_timeout * 1000}")

        if pool_size is not None and pool_size == 0:
            LOGGER.debug("No connection pool is created.")
            extra_engine_args["poolclass"] = NullPool
        else:
            if max_overflow is not None:
                extra_engine_args["max_overflow"] = max_overflow
            if pool_pre_ping is not None:
                extra_engine_args["pool_pre_ping"] = pool_pre_ping
            if pool_recycle is not None:
                extra_engine_args["pool_recycle"] = pool_recycle
            if pool_size is not None:
                extra_engine_args["pool_size"] = pool_size
            if pool_timeout is not None:
                extra_engine_args["pool_timeout"] = pool_timeout

        LOGGER.debug(f"Additional SQLAlchemy Engine args: [{extra_engine_args}]")

        engine = create_engine(connection_string, echo=False, future=True, **extra_engine_args)
        self._session_factory.configure(bind=engine)

        # Register sqlite-specific connection callback.
        if engine.dialect.name == "sqlite":
            event.listen(engine, "connect", _sqlite_on_connect)

        return engine

    def _create_or_migrate_db(self) -> None:
        """Ensure that the database schema is up to date.

        This method runs the Alembic ``upgrade`` command to ensure that all of
        the database migrations have been run and the schema is up to date.

        .. warning::

            Do not run this method concurrently against a single database, at
            risk of migrations conflicting with themselves and causing at least
            one of the callers to fail.

        """
        LOGGER.warning("Will attempt migration to latest version if needed.")

        config: Config = Config()
        config.set_main_option("script_location", str(files("buildgrid.server.sql").joinpath("alembic")))

        with self._engine.begin() as connection:
            # NOTE: pylint doesn't like this for some reason, but it is the
            # documented way to set the connection.
            # https://alembic.sqlalchemy.org/en/latest/api/config.html#alembic.config.Config
            config.attributes["connection"] = connection
            command.upgrade(config, "head")

    @property
    def dialect(self) -> str:
        """The SQL dialect in use by the configured SQL engine."""
        return self._engine.dialect.name

    @property
    def default_inlimit(self) -> int:
        """Return the default inlimit size based on the current SQL dialect"""
        return DIALECT_INLIMIT_MAP.get(self.dialect, DEFAULT_INLIMIT)

    @contextmanager
    def session(
        self,
        *,
        scoped: bool = False,
        sqlite_lock_immediately: bool = False,
        exceptions_to_not_raise_on: Optional[List[Type[Exception]]] = None,
        exceptions_to_not_rollback_on: Optional[List[Type[Exception]]] = None,
        expire_on_commit: bool = True,
    ) -> Iterator[Session]:
        """ContextManager yielding an ORM ``Session`` for the configured database.

        The :class:`sqlalchemy.orm.Session` lives for the duration of the
        managed context, and any open transaction is committed upon exiting
        the context.

        This method can potentially block for a short while before yielding
        if the underlying connection pool has recently been disposed of and
        refreshed due to connectivity issues.

        When ``sqlite_lock_immediately`` is ``True``, the Session will not
        yield until the database has been locked by entering into a write
        transaction when using SQLite.

        If an Exception is raised whilst in the managed context, the ongoing
        database transaction is rolled back, and the Exception is reraised.
        Some Exceptions which suggest a transient connection issue with the
        database lead to a ``RetriableDatabaseError`` being raised from the
        Exception instead.

        ``exceptions_to_not_raise_on`` defines a list of SQLAlchemyError types
        which should be suppressed instead of re-raised when occurring within
        the managed context.

        Similarly, ``exceptions_to_not_rollback_on`` defines a list of
        SQLAlchemyError types which will not trigger a transaction rollback
        when occuring within the managed context. Instead, the open transaction
        will be committed and the session closed.

        Args:
            scoped: If true, use a ``scoped_session`` factory to create the
                session. This results in reuse of the underlying Session object
                in a given thread.

            sqlite_lock_immediately: If true, execute a ``BEGIN IMMEDIATE``
                statement as soon as the session is created when using SQLite.
                This allows locking for the lifetime of the ``Session`` within
                this ContextManager, enabling similar behaviour to
                ``SELECT ... FOR UPDATE`` in other dialects. Defaults to
                ``False``.

            exceptions_to_not_raise_on: The list of error types to be suppressed
                within the context rather than re-raised. Defaults to ``None``,
                meaning all SQLAlchemyErrors will be re-raised.

            exceptions_to_not_rollback_on: The list
                of error types which shouldn't trigger a transaction rollback.
                Defaults to ``None``, meaning all SQLAlchemyErrors will trigger
                rollback of the transaction.

            expire_on_commit: Defaults to True. When True, all instances will
                be fully expired after each commit(), so that all attribute/object
                access subsequent to a completed transaction will load from
                the most recent database state. This flag is ignored if
                ``scoped == True``

        Yields:
            A :class:`sqlalchemy.orm.Session` object.

        Raises:
            DatabaseError: when a database session cannot be obtained.

            RetriableDatabaseError: when the database connection is temporarily
                interrupted, but can be expected to recover.

            Exception: Any Exception raised within the context will be re-raised
                unless it's type is included in the ``exceptions_to_not_raise_on``
                parameter.

        """
        if exceptions_to_not_raise_on is None:
            exceptions_to_not_raise_on = []
        if exceptions_to_not_rollback_on is None:
            exceptions_to_not_rollback_on = []

        factory: "Union[scoped_session[Session], sessionmaker[Session]]" = self._session_factory
        if scoped:
            factory = self._scoped_session_factory

        # If we recently disposed of the SQL pool due to connection issues
        # ask the client to try again when it's expected to be working again
        time_til_retry = self._sql_pool_dispose_helper.time_until_active_pool()
        if time_til_retry > timedelta(seconds=0):
            raise RetriableDatabaseError(
                "Database connection was temporarily interrupted, please retry", time_til_retry
            )

        # Try to obtain a session
        try:
            session = factory() if scoped else factory(expire_on_commit=expire_on_commit)
            if sqlite_lock_immediately and session.bind.name == "sqlite":  # type: ignore
                session.execute(text("BEGIN IMMEDIATE"))
        except Exception as e:
            LOGGER.error("Unable to obtain a database session.", exc_info=True)
            raise DatabaseError("Unable to obtain a database session.") from e

        # Yield the session and catch exceptions that occur while using it
        # to roll-back if needed
        try:
            yield session
            session.commit()
        except Exception as e:
            transient_dberr = self._sql_pool_dispose_helper.check_dispose_pool(session, e)
            if type(e) in exceptions_to_not_rollback_on:
                try:
                    session.commit()
                except Exception:
                    pass
            else:
                session.rollback()
                if transient_dberr:
                    LOGGER.warning("Rolling back database session due to transient database error.", exc_info=True)
                else:
                    LOGGER.error("Error committing database session. Rolling back.", exc_info=True)
                if type(e) not in exceptions_to_not_raise_on:
                    if transient_dberr:
                        # Ask the client to retry when the pool is expected to be healthy again
                        raise RetriableDatabaseError(
                            "Database connection was temporarily interrupted, please retry",
                            self._sql_pool_dispose_helper.time_until_active_pool(),
                        ) from e
                    raise
        finally:
            session.close()

    @contextmanager
    def scoped_session(
        self,
        *,
        sqlite_lock_immediately: bool = False,
        exceptions_to_not_raise_on: Optional[List[Type[Exception]]] = None,
        exceptions_to_not_rollback_on: Optional[List[Type[Exception]]] = None,
    ) -> Generator[Session, None, None]:
        """ContextManager providing a thread-local ORM session for the database.

        This is a shorthand for ``SqlProvider.session(scoped=True)``.

        This ContextManager provides a reusable thread-local
        :class:`sqlalchemy.orm.Session` object. Once the ``Session`` has been
        created by the initial call, subsequent calls to this method from
        within a given thread will return the same ``Session`` object until
        :meth:`SqlProvider.remove_scoped_session` is called.

        Args:
            See :meth:`SqlProvider.session` for further details.

        Yields:
            A persistent thread-local :class:`sqlalchemy.orm.Session`.

        """
        with self.session(
            scoped=True,
            sqlite_lock_immediately=sqlite_lock_immediately,
            exceptions_to_not_raise_on=exceptions_to_not_raise_on,
            exceptions_to_not_rollback_on=exceptions_to_not_rollback_on,
        ) as session:
            yield session

    def remove_scoped_session(self) -> None:
        """Remove the thread-local session, if any."""
        self._scoped_session_factory.remove()
